from __future__ import annotations

import json
import typing as t

from uniserde import Jsonable, JsonDoc

import rio

from .. import inspection, utils
from .component import Component

__all__ = [
    "FundamentalComponent",
    "KeyboardFocusableFundamentalComponent",
]


JAVASCRIPT_SOURCE_TEMPLATE = """
// Run the code in a function to avoid name clashes with globals
(function () {

    %(js_source)s


    // Make sure the user has defined the expected class.
    if (typeof %(js_user_class_name)s === 'undefined') {
        let message = `Failed to register component with unique ID \\`%(cls_unique_id)s\\` because its JavaScript source has not defined the expected class \\`%(js_user_class_name)s\\``;
        console.error(message);
        throw new Error(message);
    }


    // Wrap it in a class which inherits from Rio's ComponentBase
    class %(js_wrapper_class_name)s extends window.RIO_COMPONENT_BASE {
        createElement() {
            this.userInstance = new %(js_user_class_name)s();
            this.userInstance.__rioWrapper__ = this;
            let element = this.userInstance.createElement();
            this.userInstance.element = element;
            return element;
        }

        updateElement(deltaState, latentComponents) {
            this.userInstance.updateElement(deltaState);
        }
    }

    // Expose additional functionality to the user
    %(js_user_class_name)s.prototype.state = function () {
        return this.__rioWrapper__.state;
    }

    // Register the component
    window.COMPONENT_CLASSES['%(cls_unique_id)s'] = %(js_wrapper_class_name)s;
    window.CHILD_ATTRIBUTE_NAMES['%(cls_unique_id)s'] = %(child_attribute_names)s;
})();

"""


CSS_SOURCE_TEMPLATE = """

const style = document.createElement('style');
style.innerHTML = %(escaped_css_source)s;
document.head.appendChild(style);
"""


class FundamentalComponent(Component):
    # Unique id for identifying this class in the frontend. This is initialized
    # in `__init_subclass__`.
    _unique_id_: t.ClassVar[str]

    def build(self) -> rio.Component:
        raise RuntimeError(
            f"Attempted to call `build` on `FundamentalComponent` {self}"
        )

    @classmethod
    def build_javascript_source(cls, sess: rio.Session) -> str:
        """
        ## Metadata

        `public`: False
        """
        return ""

    @classmethod
    def build_css_source(cls, sess: rio.Session) -> str:
        """
        ## Metadata

        `public`: False
        """
        return ""

    def __init_subclass__(cls) -> None:
        # Assign a unique id to this class. This allows the frontend to identify
        # components.
        hash_ = utils.secure_string_hash(
            cls.__module__,
            cls.__qualname__,
            hash_length=12,
        )

        cls._unique_id_ = f"{cls.__name__}-{hash_}"

        # Chain up
        super().__init_subclass__()

    @classmethod
    async def _initialize_on_client(cls, sess: rio.Session) -> None:
        message_source = ""

        javascript_source = cls.build_javascript_source(sess)
        if javascript_source:
            message_source += JAVASCRIPT_SOURCE_TEMPLATE % {
                "js_source": javascript_source,
                "js_user_class_name": cls.__name__,
                "js_wrapper_class_name": f"{cls.__name__}Wrapper",
                "cls_unique_id": cls._unique_id_,
                "child_attribute_names": json.dumps(
                    inspection.get_child_component_containing_attribute_names(
                        cls
                    )
                ),
            }

        css_source = cls.build_css_source(sess)
        if css_source:
            escaped_css_source = json.dumps(css_source)
            message_source += CSS_SOURCE_TEMPLATE % {
                "escaped_css_source": escaped_css_source,
            }

        if message_source:
            await sess._evaluate_javascript(message_source)

    async def _on_message_(self, message: Jsonable, /) -> None:
        """
        This function is called when the frontend sends a message to this component
        via `sendMessage`.
        """
        raise AssertionError(
            f"Frontend sent an unexpected message to a `{type(self).__name__}`"
        )

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        """
        This function is called to check whether the delta state received from
        the frontend is valid. If the frontend tries to change the state in a
        way that isn't allowed, this function should throw an error.
        """
        raise AssertionError(
            f"Frontend tried to change the state of a `{type(self).__name__}`"
        )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        pass

    def _apply_delta_state_from_frontend(
        self, delta_state: dict[str, t.Any]
    ) -> None:
        """
        Applies the delta state received from the frontend without marking the
        component as dirty.
        """
        # Since the new state is coming from JS, we know two things:
        # 1. There's no need to send the state back to JS
        # 2. There's no need to re-build this component, since it's a
        #    FundamentalComponent
        # That means we should avoid marking this component as dirty.
        was_already_dirty = self in self.session._dirty_components

        # Update all state properties to reflect the new state
        for attr_name, attr_value in delta_state.items():
            setattr(self, attr_name, attr_value)

        if not was_already_dirty:
            self.session._dirty_components.discard(self)


class KeyboardFocusableFundamentalComponent(FundamentalComponent):
    """
    ## Metadata

    `public`: False
    """

    async def grab_keyboard_focus(self) -> None:
        """
        ## Metadata

        `public`: False
        """
        await self.session._remote_set_keyboard_focus(self._id)



from __future__ import annotations

import typing as t

import rio

from .fundamental_component import FundamentalComponent

__all__ = [
    "Overlay",
]


@t.final
class Overlay(FundamentalComponent):
    """
    Displays its content above all other components.

    The overlay component takes a single child component, and displays it above
    all other components on the page. The child will not scroll with the rest of
    the page and is exempt from layouting.

    Components inside of overlays are allocated the entire screen and are
    themselves responsible for positioning themselves as required. You can
    easily achieve this using the child's `align_x` and `align_y` properties.


    ## Attributes

    `content`: The component to display in the overlay. It will take up the
        entire size of the screen, so make sure to use properties such as
        `align_x` and `align_y` to position it as needed.


    ## Examples

    The following code will display the text "Hello, world!" centered on the
    screen:

    ```python
    rio.Overlay(
        rio.Text("Hello, world!"),
    )
    """

    content: rio.Component

    def __init__(
        self,
        content: rio.Component,
        *,
        key: str | int | None = None,
    ):
        # This component intentionally doesn't accept the typical layouting
        # parameters, as the underlying HTML `Overlay` element will force itself
        # to span the entire screen, ignoring them.

        super().__init__(key=key)

        self.content = content

    def _get_debug_details_(self) -> dict[str, t.Any]:
        result = super()._get_debug_details_()

        # Overlays intentionally remove a lot of common properties, because they
        # would behave in unexpected ways.
        del result["min_width"]
        del result["min_height"]
        del result["grow_x"]
        del result["grow_y"]
        del result["margin"]
        del result["margin_x"]
        del result["margin_y"]
        del result["margin_left"]
        del result["margin_top"]
        del result["margin_right"]
        del result["margin_bottom"]

        return result


Overlay._unique_id_ = "Overlay-builtin"



from __future__ import annotations

import dataclasses
import typing as t

from uniserde import JsonDoc

import rio

from ...components.fundamental_component import FundamentalComponent

__all__ = [
    "LayoutDisplay",
]


@t.final
class LayoutDisplay(FundamentalComponent):
    component_id: int  # This can be invalid. The component must deal with it.

    _: dataclasses.KW_ONLY

    max_requested_height: float = 1
    on_component_change: rio.EventHandler[int] = None
    on_layout_change: rio.EventHandler[[]] = None

    async def _on_message_(self, msg: t.Any) -> None:
        # Parse the message
        assert isinstance(msg, dict), msg
        assert msg["type"] == "layoutChange", msg

        # Trigger the event handler
        await self.call_event_handler(self.on_layout_change)

        # Refresh the session
        await self.session._refresh()

    def _validate_delta_state_from_frontend(self, delta_state: JsonDoc) -> None:
        if not set(delta_state) <= {"component_id"}:
            raise AssertionError(
                f"Frontend tried to change `{type(self).__name__}` state: {delta_state}"
            )

    async def _call_event_handlers_for_delta_state(
        self, delta_state: JsonDoc
    ) -> None:
        # Trigger on_change event
        try:
            new_value = delta_state["component_id"]
        except KeyError:
            pass
        else:
            assert isinstance(new_value, int), new_value
            await self.call_event_handler(self.on_component_change, new_value)


LayoutDisplay._unique_id_ = "LayoutDisplay-builtin"



from __future__ import annotations

import colorsys
import typing as t

import typing_extensions as te
from uniserde import Jsonable

import rio

from . import deprecations
from .self_serializing import SelfSerializing

__all__ = [
    "Color",
    "ColorSet",
]


def clamp_1(value: float) -> float:
    """
    Clamp a value to the range [0, 1].
    """
    if value < 0:
        return 0

    if value > 1:
        return 1

    return value


def clamp_05(value: float) -> float:
    """
    Clamp a value to the range [-0.5, 0.5].
    """
    if value < -0.5:
        return -0.5

    if value > 0.5:
        return 0.5

    return value


def _linear_rgb_to_oklab(
    r: float,
    g: float,
    b: float,
) -> tuple[float, float, float]:
    """
    Converts a color from linear sRGB to Oklab. The [original
    source](https://bottosson.github.io/posts/oklab/) is available in the public
    domain as well as the MIT license.
    """

    l = 0.4122214708 * r + 0.5363325363 * g + 0.0514459929 * b
    m = 0.2119034982 * r + 0.6806995451 * g + 0.1073969566 * b
    s = 0.0883024619 * r + 0.2817188376 * g + 0.6299787005 * b

    l_ = l ** (1 / 3)
    m_ = m ** (1 / 3)
    s_ = s ** (1 / 3)

    return (
        0.2104542553 * l_ + 0.7936177850 * m_ - 0.0040720468 * s_,
        1.9779984951 * l_ - 2.4285922050 * m_ + 0.4505937099 * s_,
        0.0259040371 * l_ + 0.7827717662 * m_ - 0.8086757660 * s_,
    )


def _oklab_to_linear_rgb(
    l: float,
    a: float,
    b: float,
) -> tuple[float, float, float]:
    """
    Converts a color from Oklab to linear sRGB. The [original
    source](https://bottosson.github.io/posts/oklab/) is available in the public
    domain as well as the MIT license.
    """

    l_ = l + 0.3963377774 * a + 0.2158037573 * b
    m_ = l - 0.1055613458 * a - 0.0638541728 * b
    s_ = l - 0.0894841775 * a - 1.2914855480 * b

    l = l_ * l_ * l_
    m = m_ * m_ * m_
    s = s_ * s_ * s_

    # Oklab can represent more colors than plain RGB. Clamping does change the
    # colors somewhat, but ensures valid RGB values.
    return (
        clamp_1(+4.0767416621 * l - 3.3077115913 * m + 0.2309699292 * s),
        clamp_1(-1.2684380046 * l + 2.6097574011 * m - 0.3413193965 * s),
        clamp_1(-0.0041960863 * l - 0.7034186147 * m + 1.7076147010 * s),
    )


@t.final
class Color(SelfSerializing):
    """
    A color, optionally with an opacity.

    The `Color` class does exactly what it says on the tin: It represents a
    single color. They're used throughout Rio to specify the color of text,
    fills, and more.

    Internally, colors are stored in the
    [Oklab](https://bottosson.github.io/posts/oklab/) color space, a modern
    color space that is designed to be perceptually uniform. Regardless of the
    internal representation, colors can be created from and converted to a
    variety of different color spaces, including linear RGB, sRGB, and HSV.

    ## Examples

    ```python
    # Color from RGB(A)
    Color.from_rgb(1.0, 0.0, 0.0, 1.0)

    # Color from hex
    Color.from_hex("#ff0000")

    # Color from HSV(A)
    Color.from_hsv(0.0, 1.0, 1.0, 1.0)

    # Color from grayscale
    Color.from_gray(0.5, 1.0)

    # Invalid: Don't call `Color` directly
    Color(1.0, 0.0, 0.0, 1.0)  # Raises a `RuntimeError`
    ```

    Regardless of how the color was created, all of the color's components will
    be accessible as attributes. For example, you can access `color.rgb`, even
    if the color was created from HSV values.


    ## Attributes

    `BLACK`: A pure black color.

    `GRAY`: A medium gray color.

    `WHITE`: A pure white color.

    `RED`: A pure red color.

    `GREEN`: A pure green color.

    `BLUE`: A pure blue color.

    `CYAN`: A pure cyan color.

    `MAGENTA`: A pure magenta color.

    `YELLOW`: A pure yellow color.

    `PINK`: A pure pink color.

    `PURPLE`: A pure purple color.

    `ORANGE`: A pure orange color.

    `BROWN`: A pure brown color.

    `TRANSPARENT`: A fully transparent color.
    """

    _l: float
    _a: float
    _b: float
    _opacity: float

    def __init__(self) -> None:
        """
        ## Metadata

        `public`: False
        """
        raise RuntimeError(
            "Don't call `Color` directly. Use `from_rgb()` and related methods instead."
        )

    @classmethod
    def from_oklab(
        cls,
        l: float,
        a: float,
        b: float,
        opacity: float = 1.0,
    ) -> Color:
        """
        Creates a color from Oklab values.

        Create a color using Oklab values. The `l` and `opacity` values must
        range from `0.0` and `1.0`, inclusive. `a` and `b` range from `-0.5` to
        `0.5`, also inclusive. This is the fastest way to instantiate a new
        `Color`, since colors are internally stored in the Oklab color space.

        ## Parameters

        `l`: The lightness of the color. `0.0` is black, `1.0` is white.

        `a`: The green-red component of the color. `-0.5` is green, `0.5` is
            red.

        `b`: The blue-yellow component of the color. `-0.5` is blue, `0.5` is
            yellow.

        `opacity`: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        `ValueError`: If any of the values are outside of their valid ranges.
        """
        # Verify all values
        if l < 0.0 or l > 1.0:
            raise ValueError("`l` must be between 0.0 and 1.0")

        if a < -0.5 or a > 0.5:
            raise ValueError("`a` must be between -0.5 and 0.5")

        if b < -0.5 or b > 0.5:
            raise ValueError("`b` must be between -0.5 and 0.5")

        if opacity < 0.0 or opacity > 1.0:
            raise ValueError("`opacity` must be between 0.0 and 1.0")

        # Instantiate the color, bypassing the blocked constructor.
        self = object.__new__(cls)

        self._l = l
        self._a = a
        self._b = b
        self._opacity = opacity

        return self

    @classmethod
    def from_rgb(
        cls,
        red: float = 1.0,
        green: float = 1.0,
        blue: float = 1.0,
        opacity: float = 1.0,
        srgb: bool = False,
    ) -> Color:
        """
        Creates a color from RGB(A) values.

        Create a color using RGB(A) values. All values must be between `0.0` and
        `1.0`, inclusive.

        If no `opacity` is given, the color will be fully opaque.

        By default, this function assumes the RGB values are in the linear RGB
        color space. To interpret them as sRGB, set the `srgb` parameter to
        `True`.

        ## Parameters

        `red`: The red component of the color. `0.0` is no red, `1.0` is full
            red.

        `green`: The green component of the color. `0.0` is no green, `1.0`
            is full green.

        `blue`: The blue component of the color. `0.0` is no blue, `1.0` is
            full blue.

        `opacity`: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        `srgb`: Whether the values are in the sRGB color space. If `True`, the
            values will be converted to linear RGB before being stored.

        ## Raises

        `ValueError`: If any of the values are outside of the range `0.0` to
            `1.0`.
        """

        # Sanity check the values
        if red < 0.0 or red > 1.0:
            raise ValueError(f"`red` must be between 0.0 and 1.0, not {red}")

        if green < 0.0 or green > 1.0:
            raise ValueError(
                f"`green` must be between 0.0 and 1.0, not {green}"
            )

        if blue < 0.0 or blue > 1.0:
            raise ValueError(f"`blue` must be between 0.0 and 1.0, not {blue}")

        if opacity < 0.0 or opacity > 1.0:
            raise ValueError(
                f"`opacity` must be between 0.0 and 1.0, not {opacity}"
            )

        # Make sure the values are linear
        if srgb:
            red = red**2.2
            green = green**2.2
            blue = blue**2.2

        # Instantiate the color
        return Color.from_oklab(
            *_linear_rgb_to_oklab(red, green, blue),
            opacity,
        )

    @classmethod
    def from_hex(
        cls,
        hex_color: str,
        srgb: bool = True,
    ) -> Color:
        """
        Parses a hex string into a color.

        Creates a color parsed from a hex string. A variety of formats are
        supported:

        - `rgb`
        - `rgba`
        - `rrggbb`
        - `rrggbbaa`

        All values may optionally be prefixed with a `#`.

        Note that unlike `from_rgb`, this function assumes the values are in the
        sRGB color space by default, as this is the most common format for hex
        colors. To interpret the values as linear RGB, set the `srgb` parameter
        to `False` instead.

        ## Parameters

        `hex_color`: The hex color string to parse.

        `srgb`: Whether the values are in the sRGB color space. If `True`, the
            values will be converted to linear RGB before being stored.

        ## Raises

        `ValueError`: If the string is not a valid hex color.
        """
        # Drop any leading `#` if present
        hex_color = hex_color.removeprefix("#")

        # Make sure the string is the correct length
        if len(hex_color) not in (3, 4, 6, 8):
            raise ValueError(
                "The hex string must be 3, 4, 6 or 8 characters long"
            )

        # Split the string into the individual components
        if len(hex_color) == 3:
            rh, gh, bh = hex_color
            ah = "f"
            max = 15
        elif len(hex_color) == 4:
            rh, gh, bh, ah = hex_color
            max = 15
        elif len(hex_color) == 6:
            rh, gh, bh = hex_color[0:2], hex_color[2:4], hex_color[4:6]
            ah = "ff"
            max = 255
        else:
            rh, gh, bh, ah = (
                hex_color[0:2],
                hex_color[2:4],
                hex_color[4:6],
                hex_color[6:8],
            )
            max = 255

        # Parse it
        return cls.from_rgb(
            red=int(rh, 16) / max,
            green=int(gh, 16) / max,
            blue=int(bh, 16) / max,
            opacity=int(ah, 16) / max,
            srgb=srgb,
        )

    @classmethod
    def from_hsv(
        cls,
        hue: float,
        saturation: float,
        value: float,
        opacity: float = 1.0,
    ) -> Color:
        """
        Create a color from HSV(A) values.

        Create a color using HSV(A) values. All values must be between `0.0` and
        `1.0`, inclusive.

        If no `opacity` is given, the color will be fully opaque.

        ## Parameters

        `hue`: The hue of the color. `0.0` is red, `0.33` is green, `0.66` is
            blue, and `1.0` is red again.

        `saturation`: The saturation of the color. `0.0` is no saturation,
            `1.0` is full saturation.

        `value`: The value of the color. `0.0` is black, `1.0` is full
            brightness.

        `opacity`: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        `ValueError`: If any of the values are outside of the range `0.0` to
            `1.0`.
        """
        if hue < 0.0 or hue > 1.0:
            raise ValueError("`hue` must be between 0.0 and 1.0")

        if saturation < 0.0 or saturation > 1.0:
            raise ValueError("`saturation` must be between 0.0 and 1.0")

        if value < 0.0 or value > 1.0:
            raise ValueError("`value` must be between 0.0 and 1.0")

        # The opacity will be checked in the `from_rgb` function, so there's no
        # need to verify it here as well.

        return cls.from_rgb(
            *colorsys.hsv_to_rgb(hue, saturation, value),
            opacity=opacity,
        )

    @classmethod
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.from_gray",
    )
    def from_grey(cls, grey: float, opacity: float = 1.0) -> Color:
        """
        Creates a grayscale color.

        Creates a gray color with the given intensity. A `grey` value of 0.0
        corresponds to black, and 1.0 to white.

        If no `opacity` is given, the color will be fully opaque.

        ## Parameters

        `grey`: The intensity of the gray color. `0.0` is black, `1.0` is
            white.

        `opacity`: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        `ValueError`: If `grey` is outside of the range `0.0` to `1.0`.
        """
        return cls.from_gray(grey, opacity)

    @classmethod
    def from_gray(cls, gray: float, opacity: float = 1.0) -> Color:
        """
        Creates a grayscale color.

        Creates a gray color with the given intensity. A `gray` value of 0.0
        corresponds to black, and 1.0 to white.

        If no `opacity` is given, the color will be fully opaque.

        ## Parameters

        `gray`: The intensity of the gray color. `0.0` is black, `1.0` is
            white.

        `opacity`: The opacity of the color. `0.0` is fully transparent,
            `1.0` is fully opaque.

        ## Raises

        `ValueError`: If `gray` is outside of the range `0.0` to `1.0`.
        """
        if gray < 0.0 or gray > 1.0:
            raise ValueError("`gray` must be between 0.0 and 1.0")

        # The opacity will be checked in the `from_oklab` function, so there's no
        # need to verify it here as well.

        return cls.from_oklab(
            l=gray,
            a=0.0,
            b=0.0,
            opacity=opacity,
        )

    @property
    def oklab(self) -> tuple[float, float, float]:
        """
        The color as Oklab values.

        The color represented as Oklab values. `l` is in range `0.0` to `1.0`
        inclusive, while `a` and `b` are in the range `-0.5` to `0.5`, also
        inclusive.
        """
        return (self._l, self._a, self._b)

    @property
    def oklaba(self) -> tuple[float, float, float, float]:
        """
        The color as Oklab values, with opacity.

        The color represented as Oklab + opacity values. `l` is in range `0.0`
        to `1.0` inclusive, while `a` and `b` are in the range `-0.5` to `0.5`,
        also inclusive.
        """
        return (self._l, self._a, self._b, self._opacity)

    @property
    def rgb(self) -> tuple[float, float, float]:
        """
        The color as linear RGB values.

        The color represented as RGB values. Each value is between `0.0` and
        `1.0`, inclusive. Note that the values are specifically in linear RGB
        space, not sRGB.
        """
        r, g, b = _oklab_to_linear_rgb(self._l, self._a, self._b)
        return (r, g, b)

    @property
    def rgba(self) -> tuple[float, float, float, float]:
        """
        The color as RGBA values.

        The color represented as RGBA values. Each value is between `0.0` and
        `1.0`, inclusive.
        """
        r, g, b = self.rgb
        return (r, g, b, self._opacity)

    @property
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.rgb[0]",
    )
    def red(self) -> float:
        """
        The color's red component.

        The red component of the color. `0.0` is no red, `1.0` is full red. Note
        that this is the linear RGB value, not sRGB.
        """
        return self.rgb[0]

    @property
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.rgb[1]",
    )
    def green(self) -> float:
        """
        The color's green component.

        The green component of the color. `0.0` is no green, `1.0` is full
        green. Note that this is the linear RGB value, not sRGB.
        """
        return self.rgb[1]

    @property
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.rgb[2]",
    )
    def blue(self) -> float:
        """
        The color's blue component.

        The blue component of the color. `0.0` is no blue, `1.0` is full blue.
        Note that this is the linear RGB value, not sRGB.
        """
        return self.rgb[2]

    @property
    def opacity(self) -> float:
        """
        How opaque the color appears.

        The opacity of the color - also know as the alpha channel. `0.0` is
        fully transparent, `1.0` is fully opaque.
        """
        return self._opacity

    @property
    def srgb(self) -> tuple[float, float, float]:
        """
        The color as sRGB values.

        The color represented as sRGB values. Each value is between `0.0` and
        `1.0`, inclusive. Note that the values are specifically in sRGB space,
        not linear RGB.
        """
        red, green, blue = self.rgb

        return (
            red ** (1 / 2.2),
            green ** (1 / 2.2),
            blue ** (1 / 2.2),
        )

    @property
    def srgba(self) -> tuple[float, float, float, float]:
        """
        The color as sRGBA values.

        The color represented as sRGBA values. Each value is between `0.0` and
        `1.0`, inclusive.
        """
        red, green, blue = self.srgb

        return (
            red,
            green,
            blue,
            self._opacity,
        )

    @property
    def hsv(self) -> tuple[float, float, float]:
        """
        The color as HSV values.

        The color represented as HSV values. Each value is between `0.0` and
        `1.0`, inclusive.
        """
        return colorsys.rgb_to_hsv(*self.rgb)

    @property
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.hsv[0]",
    )
    def hue(self) -> float:
        """
        The color's hue.

        The hue of the color, as used in the hsv color model. `0.0` is red,
        `0.33` is green, `0.66` is blue, and `1.0` is red again.
        """
        return self.hsv[0]

    @property
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.hsv[1]",
    )
    def saturation(self) -> float:
        """
        The color's saturation.

        The saturation of the color, as used in the hsv color model. `0.0` is no
        saturation, `1.0` is full saturation.
        """
        return self.hsv[1]

    @property
    @deprecations.deprecated(
        since="0.10.10",
        replacement="Color.hsv[2]",
    )
    def value(self) -> float:
        """
        The color's value.

        The value of the color, as used in the hsv color model. `0.0` is black,
        `1.0` is full brightness.
        """
        return self.hsv[2]

    @property
    def perceived_brightness(self) -> float:
        """
        How bright the color appears to humans.

        Approximates how bright the color appears to humans. `0.0` is pitch
        black, `1.0` is as bright as the screen can get.
        """
        return self._l

    @property
    def hex(self) -> str:
        """
        The color as a hex string, without opacity.

        The color, formatted as 6 hex digits. The first two digits are the red
        component, followed by the green component and finally blue. Note that
        the values are in sRGB space, as is common for hex colors.
        """
        red, green, blue = self.srgb

        red_hex = f"{int(round(red*255)):02x}"
        green_hex = f"{int(round(green*255)):02x}"
        blue_hex = f"{int(round(blue*255)):02x}"

        return red_hex + green_hex + blue_hex

    @property
    def hexa(self) -> str:
        """
        The color as a hex string, with opacity.

        The color, formatted as 8 hex digits. The first two digits are the red
        component, followed by the green component, blue, and finally the
        opacity. Note that the values are in sRGB space, as is common for hex
        colors.
        """
        rgb_hex = self.hex
        opacity_hex = f"{int(round(self._opacity*255)):02x}"

        return rgb_hex + opacity_hex

    def replace(
        self,
        *,
        red: float | None = None,
        green: float | None = None,
        blue: float | None = None,
        opacity: float | None = None,
    ) -> Color:
        """
        Replace the given values and return a new `Color` instance.

        Return a new `Color` instance with the given values replaced. Any values
        that are not given will be copied from this color.

        ## Parameters

        `red`: The red component of the new color.

        `green`: The green component of the new color.

        `blue`: The blue component of the new color.

        `opacity`: The opacity of the new color.
        """

        # TODO: What to do with the function parameters? Right now, this
        # function only allows replacing the linear RGB values, but it would be
        # much nicer to also allow replacement of the Oklab values, HSV values,
        # etc.

        cur_red, cur_green, cur_blue = self.rgb

        return Color.from_rgb(
            red=cur_red if red is None else red,
            green=cur_green if green is None else green,
            blue=cur_blue if blue is None else blue,
            opacity=self._opacity if opacity is None else opacity,
        )

    def brighter(self, amount: float) -> Color:
        """
        Return a lighter version of this color.

        Return a new `Color` instance that is brighter than this one by the
        given amount. `0` means no change, `1` will turn the color into white.
        Values less than `0` will darken the color instead.

        The function attempts to keep the hue while making the color appear
        brighter to humans. How exactly this is achieved isn't defined.

        ## Parameters

        `amount`: How much to lighten the color. `0` means no change, `1`
            will turn the color into white. Values less than `0` will
            darken the color instead.


        ## Raises

        `ValueError`: If `amount` is greater than `1`.
        """
        # The amount may be negative. If that is the case, delegate to `darker`
        if amount < 0:
            return self.darker(-amount)

        # Cannot lighten a color beyond white
        if amount > 1:
            raise ValueError("`amount` needs to be less than or equal to 1")

        # From experience, the result looks rather nonlinear. This can be
        # compensated by distorting the amount. There is no particular reason
        # for this value other than that it looks good.
        amount = amount**0.5

        # Not all colors can get equally bright. If we were to just bump the `l`
        # value towards 1 and call it a day, colors would soon go outside of the
        # range supported by RGB (and visible to humans).
        #
        # Find the maximum lightness this shade can have.
        self_hue, self_saturation, _ = self.hsv
        max_l, _, _ = Color.from_hsv(self_hue, self_saturation, 1.0).oklab

        # Interpolate `l` towards 1, making the color brighter
        brighter_l = self._l + (1 - self._l) * amount

        # TODO: This makes the color behave nonlinearly. Account for that?

        # If the resulting color is still below the maximum lightness, the
        # result is good to go.
        if brighter_l <= max_l:
            return Color.from_oklab(
                brighter_l,
                self._a,
                self._b,
                self._opacity,
            )

        # If it wasn't, find out by how much we overshot the mark
        excess = (brighter_l - max_l) / (1 - max_l)
        assert excess > 0, excess

        # Then use that to desaturate the color instead. This, combined with the
        # shift of `l` towards `1` effectively fades towards white.
        scale = 1 - excess

        return Color.from_oklab(
            l=brighter_l,
            a=self._a * scale,
            b=self._b * scale,
            opacity=self._opacity,
        )

    def darker(self, amount: float) -> Color:
        """
        Return a darker version of this color.

        Return a new `Color` instance that is darker than this one by the given
        amount. `0` means no change, `1` will turn the color into black. Values
        less than `0` will brighten the color instead.

        The function attempts to keep the hue while making the color appear
        darker to humans. How exactly this is achieved isn't defined.

        ## Parameters

        `amount`: How much to darken the color. `0` means no change, `1`
            will turn the color into black. Values less than `0` will brighten
            the color instead.

        ## Raises

        `ValueError`: If `amount` is greater than `1`.
        """
        # The value may be negative. If that is the case, delegate to `brighter`
        if amount < 0:
            return self.brighter(-amount)

        # Cannot darken a color beyond black
        if amount > 1:
            raise ValueError("`amount` needs to be less than or equal to 1")

        # HSV has an explicit value for brightness, so convert to HSV and bump
        # the value.
        #
        # The real challenge is avoiding colors from turning black. Changing a
        # 10% gray by `0.1` to black would look like an extreme change to
        # humans, even though being technically correct.
        l, a, b = self.oklab

        # Darkening by `0` should remain the same. Darkening by `1` should turn
        # any color into black. Start off with linear interpolation.
        l = l * (1 - amount)

        # TODO: This makes the color behave nonlinearly. Account for that?

        # Build the result
        return Color.from_oklab(
            l,
            a,
            b,
            self._opacity,
        )

    def desaturated(self, amount: float = 1.0) -> Color:
        """
        Returns a desaturated version of this color.

        Return a copy of this color with the saturation reduced by the given
        amount. `0` means no change, `1` will turn the color into a shade of
        gray.

        ## Parameters

        `amount`: How much to desaturate the color. `0` means no change, `1`
            will turn the color into a shade of gray.


        ## Raises

        `ValueError`: If `amount` is less than `0` or greater than `1`.
        """
        # Make sure the amount is within the valid range
        if amount < 0.0 or amount > 1.0:
            raise ValueError("`amount` must be between 0.0 and 1.0")

        # In Oklab, neutral colors are in the center at `a=0` and `b=0`, meaning
        # the coordinates just need to be scaled down.
        scale = 1 - amount

        return Color.from_oklab(
            l=self._l,
            a=self._a * scale,
            b=self._b * scale,
            opacity=self._opacity,
        )

    def blend(self, other: Color, factor: float) -> Color:
        """
        Blend this color with another color.

        Return a new `Color` instance that is a blend of this color and the
        given `other` color. `factor` controls how much of the other color is
        used. A value of `0` will return this color, a value of `1` will return
        the other color.

        Values outside of the range `0` to `1` are allowed and will lead to the
        color being extrapolated. If the resulting color would have components
        outside of their valid respective ranges, they will be clamped.

        ## Parameters

        `other`: The other color to blend with.

        `factor`: How much of the other color to use. `0` will return this
            color, `1` will return the other color.
        """
        # Plain interpolation as you'd expect. Since extrapolation is also
        # allowed, the values must be clamped to their respective valid ranges.
        one_minus_factor = 1 - factor

        return Color.from_oklab(
            l=clamp_1(
                self._l * one_minus_factor + other._l * factor,
            ),
            a=clamp_05(
                self._a * one_minus_factor + other._a * factor,
            ),
            b=clamp_05(
                self._b * one_minus_factor + other._b * factor,
            ),
            opacity=clamp_1(
                self._opacity * one_minus_factor + other._opacity * factor
            ),
        )

    @property
    def as_plotly(self) -> str:
        """
        The color as a Plotly-compatible string.

        Plotly expects colors to be specified as strings, and this function
        returns the color formatted as such.
        """
        red, green, blue, opacity = self.srgba

        return f"rgba({int(round(red*255))}, {int(round(green*255))}, {int(round(blue*255))}, {opacity})"

    def _serialize(self, sess: rio.Session) -> Jsonable:
        return self.srgba

    def __repr__(self) -> str:
        return f"<Color #{self.hexa}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Color):
            return NotImplemented

        return self.oklaba == other.oklaba

    def __hash__(self) -> int:
        return hash(self.oklaba)

    # Grays
    BLACK: t.ClassVar[Color]
    GRAY: t.ClassVar[Color]
    WHITE: t.ClassVar[Color]

    # Pure colors
    RED: t.ClassVar[Color]
    GREEN: t.ClassVar[Color]
    BLUE: t.ClassVar[Color]

    # CMY
    CYAN: t.ClassVar[Color]
    MAGENTA: t.ClassVar[Color]
    YELLOW: t.ClassVar[Color]

    # Others
    PINK: t.ClassVar[Color]
    PURPLE: t.ClassVar[Color]
    ORANGE: t.ClassVar[Color]
    BROWN: t.ClassVar[Color]

    # Special
    TRANSPARENT: t.ClassVar[Color]


Color.BLACK = Color.from_rgb(0.0, 0.0, 0.0)
Color.GREY = Color.from_rgb(0.5, 0.5, 0.5)  # type: ignore (Deprecated and thus not annotated)
Color.GRAY = Color.from_rgb(0.5, 0.5, 0.5)
Color.WHITE = Color.from_rgb(1.0, 1.0, 1.0)

Color.RED = Color.from_rgb(1.0, 0.0, 0.0)
Color.GREEN = Color.from_rgb(0.0, 1.0, 0.0)
Color.BLUE = Color.from_rgb(0.0, 0.0, 1.0)

Color.CYAN = Color.from_rgb(0.0, 1.0, 1.0)
Color.MAGENTA = Color.from_rgb(1.0, 0.0, 1.0)
Color.YELLOW = Color.from_rgb(1.0, 1.0, 0.0)

Color.PINK = Color.from_rgb(1.0, 0.0, 1.0)
Color.PURPLE = Color.from_rgb(0.5, 0.0, 0.5)
Color.ORANGE = Color.from_rgb(1.0, 0.5, 0.0)

Color.TRANSPARENT = Color.from_rgb(0.0, 0.0, 0.0, 0.0)


# Like color, but also allows referencing theme colors
ColorSet: te.TypeAlias = (
    Color
    | t.Literal[
        "background",
        "neutral",
        "hud",
        "primary",
        "secondary",
        "success",
        "warning",
        "danger",
        "keep",
    ]
)


# Cache so the session can quickly determine whether a type annotation is
# `ColorSet`
_color_set_args = set(t.get_args(ColorSet))



from __future__ import annotations

import asyncio
import functools
import inspect
import io
import itertools
import os
import sys
import tempfile
import threading
import traceback
import types
import typing as t
import warnings
import webbrowser
from datetime import timedelta
from pathlib import Path

import fastapi
import imy.docstrings
import introspection
import uvicorn
from PIL import Image

import __main__
import rio.global_state

from . import assets, global_state, maybes, routing, utils
from .app_server import fastapi_server
from .utils import ImageLike

__all__ = [
    "App",
]

P = t.ParamSpec("P")
R = t.TypeVar("R")
T = t.TypeVar("T")


def make_default_connection_lost_component() -> rio.Component:
    class DefaultConnectionLostComponent(rio.Component):
        def build(self) -> rio.Component:
            return rio.Rectangle(
                content=rio.Row(
                    rio.Icon(
                        "material/signal_disconnected",
                        fill="danger",
                        min_width=1.6,
                        min_height=1.6,
                    ),
                    rio.Text(
                        "Disconnected",
                        selectable=False,
                        style=rio.TextStyle(
                            fill=self.session.theme.hud_palette.foreground,
                            font_weight="bold",
                        ),
                    ),
                    spacing=0.8,
                    margin_x=2.5,
                    margin_y=1.5,
                ),
                fill=self.session.theme.hud_palette.background,
                corner_radius=99999,
                shadow_color=self.session.theme.shadow_color,
                shadow_radius=0.6,
                shadow_offset_y=0.15,
                margin_top=3,
                align_x=0.5,
                align_y=0.0,
            )

    return DefaultConnectionLostComponent()


def guard_against_rio_run(func: t.Callable[P, R]) -> t.Callable[P, R]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if rio.global_state.launched_via_rio_run:
            raise RuntimeError(
                f"`App.{func.__name__}` conflicts with `rio run`. Either delete"
                f' that line, or move it behind a `if __name__ == "__main__":`'
                f" guard."
            )

        return func(*args, **kwargs)

    return wrapper


@t.final
class App:
    """
    Contains all the information needed to run a Rio app.

    Apps group all the information needed for Rio to run your application, such
    as its name, icon, and the pages it contains. Apps also expose several
    lifetime events that you can use to perform tasks such as initialization and
    cleanup.

    If you're serving your app as a website, all users share the same `App`
    instance. If running in a window, there's only one window, and thus `App`,
    anyway.

    A basic setup may look like this:

    ```py
    app = rio.App(
        name="My App",
        build=MyAppRoot,
    )
    ```

    You can then run this app, either as a local application in a window:

    ```py
    app.run_in_window()
    ```

    Or you can create and run a webserver:

    ```py
    app.run_as_web_server()
    ```

    Or create a server, without running it. This allows you to start the script
    externally with tools such as uvicorn:

    ```py
    fastapi_app = app.as_fastapi()
    ```


    ## Attributes

    `name`: The name to display for this app. This can show up in window
        titles, error messages and wherever else the app needs to be
        referenced in a nice, human-readable way.

    `description`: A short, human-readable description of the app. This can
        show up in search engines, social media sites and similar.

    `pages`: The pages that make up this app. You can navigate between these
        using `Session.navigate_to` or using `Link` components. If running
        as website the user can also access these pages directly via their
        URL.

    `assets_dir`: The directory where the app's assets are stored. This allows
        you to conveniently access any images or other files that are needed
        by your app.

    `meta_tags`: Arbitrary key-value pairs that will be included in the HTML
        header of the app. These are used by search engines and social media
        sites to display information about your page, such as the title and a
        short description.
    """

    # Type hints so the documentation generator knows which fields exist
    name: str
    description: str
    assets_dir: Path
    pages: t.Sequence[rio.ComponentPage | rio.Redirect]
    meta_tags: dict[str, str]

    def __init__(
        self,
        *,
        build: t.Callable[[], rio.Component] | None = None,
        name: str | None = None,
        description: str | None = None,
        icon: ImageLike | None = None,
        pages: t.Iterable[rio.ComponentPage | rio.Redirect]
        | os.PathLike
        | str
        | None = None,
        on_app_start: rio.EventHandler[App] = None,
        on_app_close: rio.EventHandler[App] = None,
        on_session_start: rio.EventHandler[rio.Session] = None,
        on_session_close: rio.EventHandler[rio.Session] = None,
        default_attachments: t.Iterable[t.Any] = (),
        ping_pong_interval: int | float | timedelta = timedelta(seconds=50),
        assets_dir: str | os.PathLike | None = None,
        theme: rio.Theme | tuple[rio.Theme, rio.Theme] | None = None,
        build_connection_lost_message: t.Callable[
            [], rio.Component
        ] = make_default_connection_lost_component,
        meta_tags: dict[str, str] = {},
    ) -> None:
        """
        ## Parameters

        `build`: A function that returns the root component of the app. This
            function will be called whenever a new session is created. Note
            that since classes are callable in Python, you can pass a class
            here instead of a function, so long as the class doesn't require
            any arguments.

            If no build method is passed, the app will create a `PageView`
            as the root component.

        `name`: The name to display for this app. This can show up in window
            titles, error messages and wherever else the app needs to be
            referenced in a nice, human-readable way. If not specified, Rio will
            try to guess a name based on the name of the main Python file.

        `description`: A short, human-readable description of the app. This
            can show up in search engines, social media sites and similar.

        `icon`: The "favicon" to display for this app. This is a little image
            that shows up in the title bars of windows, browser tabs and similar
            similar.

        `theme`: The `Theme` for the app. You can also pass in a tuple of two
            themes, which will be used as the light mode theme and the dark mode
            theme.

        `pages`: The pages that make up this app. You can navigate between
            these using `Session.navigate_to` or using `Link` components. If
            running as website the user can also access these pages directly
            via their URL.

            Per default, rio scans your project's "pages" directory for
            components decorated with `@rio.page` and turns them into pages. To
            override the location of this directory, you can provide a custom
            path.

        `on_app_start`: A function that will be called when the app is first
            started. You can use this to perform any initialization tasks
            that need to happen before the app is ready to use.

            The app start will be delayed until this function returns. This
            makes sure initialization is complete before the app is
            displayed to the user. If you would prefer to perform
            initialization in the background try using `asyncio.create_task`
            to run your code in a separate task.

        `on_app_close`: A function that will be called right before the app
            shuts down. You can use this to clean up open resources like for
            example a database connection.

        `on_session_start`: A function that will be called each time a new
            session is created. In the context of a website that would be
            each time a new user visits the site. In the context of a window
            there is only one session, so this will only be called once.

            This function does not block the creation of the session. This
            is to make sure initialization code doesn't accidentally make
            the user wait.

            Please note that the session is not fully initialized yet when this
            function is called. In particular, the session's `active_page_url`
            is set to whichever URL the client has requested, but before the
            guards have had a chance to redirect the user to another page.

        `on_session_close`: A function that will be called each time a session
            ends. In the context of a website that would be each time a user
            closes their browser tab. In the context of a window this will
            only be called once, when the window is closed.

        `default_attachments`: A list of attachments that will be attached to
            every new session. (All sessions share the exact same attachment
            objects, no copies are created.)

        `ping_pong_interval`: Rio periodically sends ping-pong messages
            between the client and server to prevent overzealous proxies
            from closing the connection. The default value should be fine
            for most deployments, but feel free to change it if your hosting
            provider deploys a particularly obnoxious proxy.

        `assets_dir`: The directory where the app's assets are stored. This
            allows you to conveniently access any images or other files that
            are needed by your app. If not specified, Rio will assume the
            assets are stored in a directory called "assets" in the same
            directory as the main Python file.

        `build_connection_lost_message`: A function that creates a "Connection
            lost" error popup, in case you want to override the default one.

        `meta_tags`: Arbitrary key-value pairs that will be included in the
            HTML header of the app. These are used by search engines and social
            media sites to display information about your page, such as the
            title and a short description.
        """
        # A common mistake is to pass types instead of instances to
        # `default_attachments`. Catch that, scream and die.
        for attachment in default_attachments:
            if isinstance(attachment, type):
                raise TypeError(
                    f"Default attachments should be instances, not types. Did you mean to type `{attachment.__name__}()`?"
                )

        if description is None:
            description = "A Rio web-app written in 100% Python"

        if icon is None:
            icon = utils.HOSTED_ASSETS_DIR / "rio_logos/rio_logo_square.png"

        if build is None:
            build = rio.components.default_root_component.DefaultRootComponent

        if theme is None:
            theme = rio.Theme.from_colors()

        if name is None:
            name = self._infer_app_name()

        if assets_dir is None:
            assets_dir = self._base_dir_for_relative_paths / "assets"
        else:
            assets_dir = self._base_dir_for_relative_paths / assets_dir

        if pages is None:
            pages = self._infer_pages()
        elif isinstance(pages, (os.PathLike, str)):
            pages = routing.auto_detect_pages(
                self._base_dir_for_relative_paths / pages
            )
        else:
            pages = list(pages)

        self.name = name
        self.description = description
        self.assets_dir = assets_dir
        self.pages = pages
        self._build = build
        self._icon = assets.Asset.from_image(icon)
        self._on_app_start = on_app_start
        self._on_app_close = on_app_close
        self._on_session_start = on_session_start
        self._on_session_close = on_session_close
        self.default_attachments = list(default_attachments)
        self._theme = theme
        self._build_connection_lost_message = build_connection_lost_message
        self._custom_meta_tags = meta_tags

        if isinstance(ping_pong_interval, timedelta):
            self._ping_pong_interval = ping_pong_interval
        else:
            self._ping_pong_interval = timedelta(seconds=ping_pong_interval)

        # Initialized lazily, when the icon is first requested
        #
        # This starts out as `None`, then either becomes a `bytes` object if
        # it's successfully fetched, or an error message if fetching failed.
        self._icon_as_png_blob: bytes | str | None = None

        # All extensions currently registered with the app, by their `id()`s
        self._ids_to_extensions: dict[int, rio.Extension] = {}

        # These keep track of functions which must be called at certain points.
        # They are collected into lists here so they can be called quickly.
        self._extension_on_as_fastapi_handlers: list[
            t.Callable[[rio.ExtensionAsFastapiEvent], None]
        ] = []

        self._extension_on_app_start_handlers: list[
            t.Callable[[rio.ExtensionAppStartEvent], None]
        ] = []
        self._extension_on_app_close_handlers: list[
            t.Callable[[rio.ExtensionAppCloseEvent], None]
        ] = []

        self._extension_on_session_start_handlers: list[
            t.Callable[[rio.ExtensionSessionStartEvent], None]
        ] = []
        self._extension_on_session_close_handlers: list[
            t.Callable[[rio.ExtensionSessionCloseEvent], None]
        ] = []

        self._extension_on_page_change_handlers: list[
            t.Callable[[rio.ExtensionPageChangeEvent], None]
        ] = []

    async def _call_event_handlers(
        self,
        handlers: t.Iterable[t.Callable[[T], t.Any | t.Awaitable[t.Any]]],
        event_data: T,
    ) -> None:
        """
        Calls all event handlers in the given list, guarding against exceptions.
        """

        for handler in handlers:
            # If the handler is available, call it and await it if necessary
            try:
                result = handler(event_data)

                if inspect.isawaitable(result):
                    await result

            # Display and discard exceptions
            except Exception:
                print("Exception in event handler:")
                traceback.print_exc()

    def _call_event_handlers_sync(
        self,
        handlers: t.Iterable[t.Callable[[T], t.Any]],
        event_data: T,
    ) -> None:
        """
        Same as `_call_event_handlers`, but without support for asynchronous
        handlers.
        """

        for handler in handlers:
            # If the handler is available, call it and await it if necessary
            try:
                handler(event_data)

            # Display and discard exceptions
            except Exception:
                print("Exception in event handler:")
                traceback.print_exc()

    @imy.docstrings.mark_as_private
    async def fetch_icon_png_blob(self) -> bytes:
        """
        Fetches the app's icon as a PNG blob.

        The result is cached. It will be loaded the first time you call this
        method, and then returned immediately on subsequent calls. If fetching
        the icon fails, the exception is also cached, and no further fetching
        attempts will be made.


        ## Raises

        `IOError`: If the icon could not be fetched.
        """

        # Already cached?
        if isinstance(self._icon_as_png_blob, bytes):
            return self._icon_as_png_blob

        # Already failed?
        if isinstance(self._icon_as_png_blob, str):
            raise IOError(self._icon_as_png_blob)

        # Nope, get it
        try:
            icon_blob, _ = await self._icon.try_fetch_as_blob()

            input_buffer = io.BytesIO(icon_blob)
            output_buffer = io.BytesIO()

            with Image.open(input_buffer) as image:
                image.save(output_buffer, format="png")

        except Exception as err:
            if isinstance(self._icon, assets.PathAsset):
                message = f"Could not fetch the app's icon from {self._icon.path.absolute()}"
            elif isinstance(self._icon, assets.UrlAsset):
                message = (
                    f"Could not fetch the app's icon from {self._icon.url}"
                )
            else:
                message = "Could not fetch the app's icon"

            self._icon_as_png_blob = message
            raise IOError(message) from err

        # Cache it
        self._icon_as_png_blob = output_buffer.getvalue()

        # Done!
        return self._icon_as_png_blob

    @imy.docstrings.mark_as_private
    async def fetch_icon_as_png_path(self) -> Path | None:
        """
        Fetches the app's icon and returns the path to it, as PNG file. This
        will take care of fetching it (if needed) and converting it to PNG.

        If the icon file isn't local, it will be stored in a temporary
        directory. Note that since the result isn't a context manager, the file
        won't be deleted.

        If the icon can't be fetched, a warning is displayed and `None` is
        returned.
        """
        try:
            # If the icon is a local PNG file, use it directly
            if (
                isinstance(self._icon, assets.PathAsset)
                and self._icon.path.suffix == ".png"
            ):
                if self._icon.path.exists():
                    return self._icon.path
                else:
                    raise FileNotFoundError(
                        f"{self._icon.path!r} does not exist"
                    )

            # Otherwise fetch it
            png_blob = await self.fetch_icon_png_blob()

            # Dump it to a temporary file
            with tempfile.NamedTemporaryFile(
                delete=False, suffix=".png"
            ) as file:
                file.write(png_blob)

            return Path(file.name)
        except IOError as error:
            print("aairtnsiarnsoin", error)
            warnings.warn(
                f"Failed to load app icon: {type(error).__name__} {error}"
            )
            return None

    @functools.cached_property
    def _main_file_path(self) -> Path:
        if global_state.rio_run_app_module_path is not None:
            main_file = global_state.rio_run_app_module_path
        else:
            try:
                main_file = Path(__main__.__file__)
            except AttributeError:
                main_file = Path(sys.argv[0])

        # Find out if we're being executed by uvicorn
        if (
            main_file.name != "__main__.py"
            or main_file.parent != Path(uvicorn.__file__).parent
        ):
            return main_file

        # Find out from which module uvicorn imported the app
        try:
            app_location = next(arg for arg in sys.argv[1:] if ":" in arg)
        except StopIteration:
            return main_file

        module_name, _, _ = app_location.partition(":")
        module = sys.modules[module_name]

        if module.__file__ is not None:
            return Path(module.__file__)

        return main_file

    @functools.cached_property
    def _calling_module(self) -> types.ModuleType:
        call_frame = introspection.CallFrame.current()

        while call_frame is not None:
            module_name = call_frame.globals["__name__"]

            if module_name.startswith("rio.") or module_name == "functools":
                call_frame = call_frame.parent
                continue

            return sys.modules[module_name]

        assert False, "Calling module could not be found?!"

    @functools.cached_property
    def _base_dir_for_relative_paths(self) -> Path:
        if self._calling_module.__file__ is None:
            # TODO: What do we do in this case?
            return Path.cwd()

        return Path(self._calling_module.__file__).parent

    def _infer_app_name(self) -> str:
        main_file_path = self._main_file_path

        name = main_file_path.stem
        if name in ("main", "__main__", "__init__"):
            name = main_file_path.absolute().parent.name

        return name.replace("_", " ").title()

    def _infer_pages(self) -> list[rio.ComponentPage]:
        pages_dir = self._base_dir_for_relative_paths / "pages"

        # No `pages` folder exists? No pages, then.
        if not pages_dir.exists():
            return []

        # In order to import the pages correctly correctly we must know the
        # correct module name to use. We know that the `pages` folder is a
        # sibling of `self._calling_module`, so we can deduce the correct module
        # name based on that.
        if self._calling_module.__file__ is None:
            # TODO: Display a warning?
            return []

        caller_file_path = Path(self._calling_module.__file__)

        # Careful: If the calling file is the `__init__.py` of a package, the
        # `__init__` part is not included in the module's `__name__`.
        if (
            caller_file_path.name == "__init__.py"
            and self._calling_module.__name__.rpartition(".")[2] != "__init__"
        ):
            package_name = self._calling_module.__name__
        else:
            package_name = self._calling_module.__name__.rpartition(".")[0]

        if package_name:
            package_name += ".pages"
        else:
            package_name = "pages"

        return routing.auto_detect_pages(pages_dir, package=package_name)

    def _iter_page_urls(
        self,
        *,
        include_redirects: bool,
    ) -> t.Iterator[str]:
        """
        Generates valid page URLs for this app. If the possible values for
        placeholders are known (such as with `Literal`s) then all possible
        combinations are generated. Otherwise, the URL is skipped.

        Can be used to generate a sitemap, test that no pages crash, etc.

        ## Parameters

        `include_redirects`: Whether to include URLs for `rio.Redirect` pages.
        """

        def urls_for_pages(
            pages: t.Iterable[rio.ComponentPage | rio.Redirect],
        ) -> t.Iterator[str]:
            for page in pages:
                if isinstance(page, rio.ComponentPage):
                    yield from urls_for_component_page(page)

                elif include_redirects and isinstance(page, rio.Redirect):
                    yield page.url_segment

        def urls_for_component_page(page: rio.ComponentPage) -> t.Iterator[str]:
            # A page can have multiple URLs if it has path parameters. Start by
            # creating a list of URLs that are valid for this page.
            if page._url_pattern.path_parameter_names:
                # If this page has path parameters, we don't know what it
                # considers a valid URL. One exception are Literals - we simply
                # insert every allowed value into the URL.
                options = dict[str, t.Sequence[str]]()

                for parameter in page._url_pattern.path_parameter_names:
                    parser = page._url_parameter_parsers[parameter]

                    try:
                        options[parameter] = parser.list_valid_values()
                    except ValueError:
                        # If a parser can't reasonably list all of its valid
                        # values, then we simply won't generate any URLs for
                        # this page at all.
                        return

                urls = (
                    page._url_pattern.build_url(dict(zip(options, combination)))
                    for combination in itertools.product(*options.values())
                )
            else:
                urls = [page.url_segment]

            # Now that we know all the URLs for *this* page, we need the urls
            # for the sub-pages.
            #
            # There is a distinction here: If a page has children, then the page
            # url *must* be combined with a child url. For example, if `/foo`
            # has a sub-page `/bar`, then `/foo/bar` is a valid URL, but `/foo`
            # is not! (Unless foo also has a sub-page with `url_segment=""`.)
            if not page.children:
                yield from urls
                return

            child_urls = [
                "/" + child_url for child_url in urls_for_pages(page.children)
            ]

            # Finally, combine all of our own URLs with the URLs of the child
            # pages
            for url in urls:
                for child_url in child_urls:
                    yield url + child_url

        # Remove trailing slashes. (This has no functional purpose, we just
        # don't want to output redundant slashes.)
        return (url.rstrip("/") for url in urls_for_pages(self.pages))

    def _as_fastapi(
        self,
        *,
        debug_mode: bool,
        running_in_window: bool,
        internal_on_app_start: t.Callable[[], t.Any] | None,
        base_url: rio.URL | str | None,
    ) -> fastapi.FastAPI:
        """
        Internal equivalent of `as_fastapi` that takes additional arguments.
        """
        # Make sure all globals are initialized. This should be done as late as
        # possible, because it depends on which modules have been imported into
        # `sys.modules`.
        maybes.initialize()

        # For convenience, this method can accept a string as the base URL.
        # Convert that
        if isinstance(base_url, str):
            base_url = rio.URL(base_url)

        # Build the fastapi instance
        result = fastapi_server.FastapiServer(
            self,
            debug_mode=debug_mode,
            running_in_window=running_in_window,
            internal_on_app_start=internal_on_app_start,
            base_url=base_url,
        )

        # Call all extension event handlers
        self._call_event_handlers_sync(
            self._extension_on_as_fastapi_handlers,
            rio.ExtensionAsFastapiEvent(
                self,
                result,
            ),
        )

        return result

    def as_fastapi(
        self,
        *,
        base_url: rio.URL | str | None = None,
    ) -> fastapi.FastAPI:
        """
        Return a FastAPI instance that serves this app.

        This method returns a FastAPI instance that serves this app. This allows
        you to run the app with a custom server, such as uvicorn:

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        fastapi_app = app.as_fastapi()
        ```

        You can then run this app via `uvicorn`:

        ```sh
        uvicorn my_app:fastapi_app
        ```

        ## Parameters

        `base_url`: The base URL at which the app will be served. This is useful
            if you're running the app behind a reverse proxy like nginx and want
            to serve the app at a subpath. If provided, the URL must be absolute
            and cannot contain query parameters or fragments.

            **This parameter is experimental. Please report any issues you
            encounter. Minor releases may change the behavior of this
            parameter.**
        """
        return self._as_fastapi(
            debug_mode=False,
            running_in_window=False,
            internal_on_app_start=None,
            base_url=base_url,
        )

    def _run_as_web_server(
        self,
        *,
        host: str,
        port: int,
        quiet: bool,
        running_in_window: bool,
        internal_on_app_start: t.Callable[[], None] | None = None,
        internal_on_server_created: t.Callable[[uvicorn.Server], None]
        | None = None,
        base_url: rio.URL | str | None = None,
    ) -> None:
        """
        Internal equivalent of `run_as_web_server` that takes additional
        arguments.
        """
        port = utils.ensure_valid_port(host, port)

        # Suppress stdout messages if requested
        kwargs = {}

        if quiet:
            kwargs["log_config"] = {
                "version": 1,
                "disable_existing_loggers": True,
                "formatters": {},
                "handlers": {},
                "loggers": {},
            }

        # Create the FastAPI server
        fastapi_app = self._as_fastapi(
            debug_mode=False,
            running_in_window=running_in_window,
            internal_on_app_start=internal_on_app_start,
            base_url=base_url,
        )

        # Suppress stdout messages if requested
        log_level = "error" if quiet else "info"

        config = uvicorn.Config(
            fastapi_app,
            host=host,
            port=port,
            log_level=log_level,
            timeout_graceful_shutdown=1,  # Without a timeout, sometimes the server just deadlocks
        )
        server = uvicorn.Server(config)

        if internal_on_server_created is not None:
            internal_on_server_created(server)

        server.run()

    @guard_against_rio_run
    def run_as_web_server(
        self,
        *,
        host: str = "localhost",
        port: int = 8000,
        quiet: bool = False,
        base_url: rio.URL | str | None = None,
    ) -> None:
        """
        Creates and runs a webserver that serves this app.

        This method creates and immediately runs a webserver that serves this
        app. This is the simplest way to run a Rio app.

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        app.run_as_web_server()
        ```

        The will synchronously block until the server is shut down.

        ## Parameters

        host: Which IP address to serve the webserver on. `localhost` will
            make the service only available on your local machine. This is
            the recommended setting if running behind a proxy like nginx.

        port: Which port the webserver should listen to.

        quiet: If `True` Rio won't send any routine messages to `stdout`.
            Error messages will be printed regardless of this setting.

        `base_url`: The base URL at which the app will be served. This is useful
            if you're running the app behind a reverse proxy like nginx and want
            to serve the app at a subpath. If provided, the URL must be absolute
            and cannot contain query parameters or fragments.

            **This parameter is experimental. Please report any issues you
            encounter. Minor releases may change the behavior of this
            parameter.**
        """
        self._run_as_web_server(
            host=host,
            port=port,
            quiet=quiet,
            running_in_window=False,
            base_url=base_url,
        )

    @guard_against_rio_run
    def run_in_browser(
        self,
        *,
        host: str = "localhost",
        port: int | None = None,
        quiet: bool = False,
    ) -> None:
        """
        Runs an internal webserver and opens the app in the default browser.

        This method creates and immediately runs a webserver that serves this
        app, and then opens the app in the default browser. This is a quick and
        easy way to access your app.

        ```py
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        app.run_in_browser()
        ```

        ## Parameters
        host: Which IP address to serve the webserver on. `localhost` will
            make the service only available on your local machine. This is the
            recommended setting if running behind a proxy like nginx.

        port: Which port the webserver should listen to. If not specified,
            Rio will choose a random free port.

        quiet: If `True` Rio won't send any routine messages to `stdout`.
            Error messages will be printed regardless of this setting.
        """
        port = utils.ensure_valid_port(host, port)

        def on_startup() -> None:
            webbrowser.open(f"http://{host}:{port}")

        self._run_as_web_server(
            host=host,
            port=port,
            quiet=quiet,
            running_in_window=False,
            internal_on_app_start=on_startup,
        )

    @guard_against_rio_run
    def run_in_window(
        self,
        *,
        quiet: bool = True,
        maximized: bool = False,
        fullscreen: bool = False,
        width: float | None = None,
        height: float | None = None,
    ) -> None:
        """
        Runs the app in a local window.

        This method creates a window and displays the app in it. This is great
        if you don't want the complexity of running a web server, or wish to
        package your app as a standalone executable.

        ```python
        app = rio.App(
            name="My App",
            build=MyAppRoot,
        )

        app.run_in_window()
        ```

        This method requires the `window` extra. If you don't have it installed,
        you can install it with:

        ```sh
        pip install "rio-ui[window]"
        ```

        This method will synchronously block until the window is closed.


        ## Parameters

        `quiet`: If `True` Rio won't send any routine messages to `stdout`.
            Error messages will be printed regardless of this setting.

        `maximized`: Whether to maximize the app window.

        `fullscreen`: Whether to open the app window in fullscreen mode.

        `width`: The default width of the app window.

        `height`: The default height of the app window.
        """
        try:
            from . import webview_shim
        except ImportError:
            raise Exception(
                "The `window` extra is required to use `App.run_in_window`."
                """ Run `pip install "rio-ui[window]"` to install it."""
            ) from None

        # Unfortunately, WebView must run in the main thread, which makes this
        # tricky. We'll have to banish uvicorn to a background thread, and shut
        # it down when the window is closed.

        host = "localhost"
        port = utils.ensure_valid_port(host, None)
        url = f"http://{host}:{port}"

        # This lock is released once the server is running
        app_ready_event = threading.Event()

        server: uvicorn.Server | None = None

        def on_server_created(serv: uvicorn.Server) -> None:
            nonlocal server
            server = serv

        # Start the server, and release the lock once it's running
        def run_web_server() -> None:
            self._run_as_web_server(
                host=host,
                port=port,
                quiet=quiet,
                running_in_window=True,
                internal_on_app_start=app_ready_event.set,
                internal_on_server_created=on_server_created,
            )

        server_thread = threading.Thread(target=run_web_server)
        server_thread.start()

        # Wait for the server to start
        app_ready_event.wait()

        # Problem: width and height are given in rem, but we need them in
        # pixels. We'll use pywebview's evaluate_js to find out as soon as the
        # window has been created, and then update the window size accordingly.
        def update_window_size() -> None:
            if width is None and height is None:
                return

            pixels_per_rem = window.evaluate_js("""
let measure = document.createElement('div');
measure.style.height = '1rem';

let pixels_per_rem = measure.getBoundingClientRect().height * window.devicePixelRatio;

measure.remove();

pixels_per_rem;
""")

            if width is None:
                width_in_pixels = window.width
            else:
                width_in_pixels = round(width * pixels_per_rem)

            if height is None:
                height_in_pixels = window.height
            else:
                height_in_pixels = round(height * pixels_per_rem)

            window.set_window_size(width_in_pixels, height_in_pixels)

        # Fetch the icon
        icon_path = asyncio.run(self.fetch_icon_as_png_path())

        # Start the webview
        try:
            window = webview_shim.create_window(
                self.name,
                url,
                maximized=maximized,
                fullscreen=fullscreen,
            )
            webview_shim.start_mainloop(
                update_window_size,
                icon=icon_path,
            )

        finally:
            assert isinstance(server, uvicorn.Server)

            server.should_exit = True
            server_thread.join()

    def _add_extension(self, /, extension: rio.Extension) -> None:
        """
        Registers an extension with the app.

        This function adds an extension to the app, giving the extension access
        to the app's events.

        ## Raises

        `ValueError`: If this extension is already registered with the app.
        """
        # Make sure this particular extension isn't already registered
        if id(extension) in self._ids_to_extensions:
            raise ValueError(
                f"Extension {extension} is already registered with the app"
            )

        # Register the extension
        self._ids_to_extensions[id(extension)] = extension

        # Gather all of the the extension's event handlers. This will put them
        # in a dictionary, grouped by their event tag.
        handlers = rio.extension_event._collect_tagged_methods_recursive(
            extension
        )

        # The values in the dictionary above aren't just the callables - they
        # allow for an optional argument that was passed to the decorator. Since
        # most events don't use that, use a helper function to strip it again.
        def extend_with_first_in_tuples(target: list, tuples) -> None:
            for tup in tuples:
                assert len(tup) == 2
                assert tup[1] is None

                target.append(tup[0])

        extend_with_first_in_tuples(
            self._extension_on_as_fastapi_handlers,
            handlers.get(
                rio.extension_event.ExtensionEventTag.ON_AS_FASTAPI, []
            ),
        )

        extend_with_first_in_tuples(
            self._extension_on_app_start_handlers,
            handlers.get(
                rio.extension_event.ExtensionEventTag.ON_APP_START, []
            ),
        )
        extend_with_first_in_tuples(
            self._extension_on_app_close_handlers,
            handlers.get(
                rio.extension_event.ExtensionEventTag.ON_APP_CLOSE, []
            ),
        )

        extend_with_first_in_tuples(
            self._extension_on_session_start_handlers,
            handlers.get(
                rio.extension_event.ExtensionEventTag.ON_SESSION_START, []
            ),
        )
        extend_with_first_in_tuples(
            self._extension_on_session_close_handlers,
            handlers.get(
                rio.extension_event.ExtensionEventTag.ON_SESSION_CLOSE, []
            ),
        )

        extend_with_first_in_tuples(
            self._extension_on_page_change_handlers,
            handlers.get(
                rio.extension_event.ExtensionEventTag.ON_PAGE_CHANGE, []
            ),
        )

    def add_default_attachment(self, attachment: t.Any, /) -> None:
        """
        Adds a new default attachment to the app.

        Default attachments are automatically attached to every new session
        created by the app. This is useful for providing shared resources to
        all sessions, such as database connections or other global state.

        This function adds a new default attachment to the app, causing it to
        be attached to every new session.

        ## Parameters

        `attachment`: The value to attach to the session. You can retrieve it
            later by its type.

        ## Raises

        `TypeError`: If the attachment is a class instead of an instance of a
            class.
        """
        if isinstance(attachment, type):
            raise TypeError(
                f"Default attachments should be instances, not types. Did you mean to type `add_default_attachment({attachment.__name__}())`?"
            )

        self.default_attachments.append(attachment)

    def __getitem__(self, key: t.Type[T], /) -> T:
        """
        Retrieves a default attachment by its type.

        Default attachments are automatically attached to every new session
        created by the app. This is useful for providing shared resources to
        all sessions, such as database connections or other global state.

        This function retrieves a default attachment from this app. To attach
        values to the session, use `App.add_default_attachment`.

        ## Parameters

        `key`: The class of the value you want to retrieve.

        ## Raises

        `KeyError`: If no attachment of this type is attached to the session.
        """
        for attachment in self.default_attachments:
            if isinstance(attachment, key):
                return attachment

        raise KeyError(key)



import functools
import io
import typing as t

import rio


def colors_equal(color1: rio.Color, color2: rio.Color) -> bool:
    """
    Returns `True` if the two colors are equal and `False` otherwise. Since
    color values are stored as floats, this function applies a small fudge
    factor to account for floating point differences.
    """
    epsilon = 1e-6

    l1, a1, b1, opacity1 = color1.oklaba
    l2, a2, b2, opacity2 = color2.oklaba

    return (
        abs(l1 - l2) < epsilon
        and abs(a1 - a2) < epsilon
        and abs(b1 - b2) < epsilon
        and abs(opacity1 - opacity2) < epsilon
    )


def get_minimum_theme_kwargs(theme: rio.Theme) -> dict[str, t.Any]:
    """
    Given a theme, returns a dictionary with the minimum set of keyword
    arguments required to recreate it.
    """
    # This is more complex than it might seem at first, because many colors are
    # derived from other colors. For example, the neutral color is derived from
    # the primary one.
    result: dict[str, t.Any] = {}

    # Light / dark mode can impact some colors. Make sure to get that value
    # first.
    if not theme.is_light_theme:
        result["mode"] = "dark"

    # Some colors don't depend on anything else
    reference_theme = rio.Theme.from_colors(**result)

    if not colors_equal(theme.primary_color, reference_theme.primary_color):
        result["primary_color"] = theme.primary_color

    if not colors_equal(theme.secondary_color, reference_theme.secondary_color):
        result["secondary_color"] = theme.secondary_color

    if not colors_equal(theme.disabled_color, reference_theme.disabled_color):
        result["disabled_color"] = theme.disabled_color

    if not colors_equal(theme.success_color, reference_theme.success_color):
        result["success_color"] = theme.success_color

    if not colors_equal(theme.warning_color, reference_theme.warning_color):
        result["warning_color"] = theme.warning_color

    if not colors_equal(theme.danger_color, reference_theme.danger_color):
        result["danger_color"] = theme.danger_color

    if not colors_equal(theme.hud_color, reference_theme.hud_color):
        result["hud_color"] = theme.hud_color

    # These depend on the previously defined ones
    reference_theme = rio.Theme.from_colors(**result)

    if not colors_equal(
        theme.background_color, reference_theme.background_color
    ):
        result["background_color"] = theme.background_color

    if not colors_equal(theme.neutral_color, reference_theme.neutral_color):
        result["neutral_color"] = theme.neutral_color

    # Header fill
    #
    # This is nontrivial, because there are many kinds of fill, and some of them
    # can be hard to serialize. Only support solid colors for now.
    heading_color = theme.heading1_style.fill
    reference_heading_color = reference_theme.heading1_style.fill

    assert isinstance(heading_color, rio.Color), heading_color
    assert isinstance(
        reference_heading_color, rio.Color
    ), reference_heading_color

    if isinstance(heading_color, rio.Color) and not colors_equal(
        heading_color, reference_heading_color
    ):
        result["heading_fill"] = heading_color

    # Corner radii
    if theme.corner_radius_large != reference_theme.corner_radius_large:
        result["corner_radius_large"] = theme.corner_radius_large

    if theme.corner_radius_medium != reference_theme.corner_radius_medium:
        result["corner_radius_medium"] = theme.corner_radius_medium

    if theme.corner_radius_small != reference_theme.corner_radius_small:
        result["corner_radius_small"] = theme.corner_radius_small

    return result


async def update_and_apply_theme(
    session: rio.Session,
    theme_replacements: dict[str, t.Any],
) -> None:
    """
    Overrides the session's theme with the given one, and makes sure to update
    all components so they use the new theme.
    """

    # Determine the kwargs to use for the theme
    theme_kwargs = get_minimum_theme_kwargs(session.theme)
    theme_kwargs.update(theme_replacements)

    # Build the new theme
    new_theme = rio.Theme.from_colors(**theme_kwargs)

    # Apply it
    await session._apply_theme(new_theme)

    # The application itself isn't enough, because some components will have
    # read theme values and used them to set e.g. their corner radii. Dirty
    # every component to force a full rebuild.
    for component in session._weak_components_by_id.values():
        session._register_dirty_component(
            component,
            include_children_recursively=False,
        )

    # Refresh
    await session._refresh()


def get_source_for_theme(theme: rio.Theme, *, create_theme_pair: bool) -> str:
    """
    Given a theme, returns a string that can be used to recreate it.
    """
    # Find all parameters which must be passed to create this theme
    theme_parameters = get_minimum_theme_kwargs(theme)

    # Build the source
    theme_or_themes = "themes" if create_theme_pair else "theme"
    result = io.StringIO()
    result.write(f"# Create the {theme_or_themes}\n")

    if create_theme_pair:
        result.write("themes = rio.Theme.pair_from_colors(")
    else:
        result.write("theme = rio.Theme.from_colors(")

    if theme_parameters:
        result.write("\n")

        for key, value in theme_parameters.items():
            result.write(f"    {key}=")

            if isinstance(value, rio.Color):
                hex_value = value.hexa
                if len(hex_value) == 8 and hex_value.endswith("ff"):
                    hex_value = hex_value[:-2]

                result.write(f"rio.Color.from_hex({hex_value!r})")
            elif isinstance(value, bool):
                result.write("True" if value else "False")
            elif isinstance(value, (int, float)):
                result.write(f"{value:.2f}")
            elif isinstance(value, str):
                result.write(repr(value))
            else:
                raise NotImplementedError(f"Unsupported type: {type(value)}")

            result.write(",\n")

    result.write(")\n")
    result.write("\n")
    result.write(
        f"# And apply {'them' if create_theme_pair else 'it'} to your app\n"
    )
    result.write("app = rio.App(\n")
    result.write("    ...\n")
    result.write(f"    theme={theme_or_themes},\n")
    result.write("    ...\n")
    result.write(")")

    # Done
    return result.getvalue()


class PalettePicker(rio.Component):  #
    shared_open_key: str

    palette_nicename: str
    palette_slug: str

    pick_opacity: bool = False

    round_top: bool = False
    round_bottom: bool = False

    @property
    def palette(self) -> rio.Palette:
        return getattr(self.session.theme, f"{self.palette_slug}_palette")

    async def _on_color_change(self, event: rio.ColorChangeEvent) -> None:
        await update_and_apply_theme(
            self.session,
            {
                f"{self.palette_slug}_color": event.color,
            },
        )

    def _on_press(self, event: rio.PointerEvent) -> None:
        # Toggle the popup
        if self.shared_open_key == self.palette_nicename:
            self.shared_open_key = ""
        else:
            self.shared_open_key = self.palette_nicename

    def build(self) -> rio.Component:
        palette = self.palette

        top_radius = (
            self.session.theme.corner_radius_medium if self.round_top else 0
        )
        bottom_radius = (
            self.session.theme.corner_radius_medium if self.round_bottom else 0
        )

        return rio.Popup(
            anchor=rio.PointerEventListener(
                # Switches the color of the Rectangle's ripple effect
                rio.ThemeContextSwitcher(
                    content=rio.Rectangle(
                        content=rio.Column(
                            rio.Text(
                                self.palette_nicename,
                                style=rio.TextStyle(
                                    font_size=self.session.theme.heading3_style.font_size,
                                    fill=palette.foreground,
                                ),
                                selectable=False,
                                justify="left",
                            ),
                            rio.Text(
                                f"#{palette.background.hexa}",
                                style=rio.TextStyle(
                                    font_size=1,
                                    fill=palette.foreground.replace(
                                        opacity=0.5
                                    ),
                                ),
                                justify="left",
                            ),
                            spacing=0.2,
                            margin_x=1,
                            margin_y=0.8,
                        ),
                        fill=palette.background,
                        corner_radius=(
                            top_radius,
                            top_radius,
                            bottom_radius,
                            bottom_radius,
                        ),
                        ripple=True,
                        cursor=rio.CursorStyle.POINTER,
                        transition_time=0.15,
                    ),
                    color=palette.background,
                ),
                on_press=self._on_press,
            ),
            content=rio.Column(
                rio.Text(
                    f"{self.palette_nicename} Color",
                    justify="center",
                    style="heading3",
                ),
                rio.ColorPicker(
                    color=palette.background,
                    pick_opacity=self.pick_opacity,
                    on_change=self._on_color_change,
                    min_width=18,
                    min_height=16,
                ),
                spacing=0.8,
                margin=1,
            ),
            is_open=self.shared_open_key == self.palette_nicename,
            color="hud",
            position="left",
            gap=1,
        )


class ThemePickerPage(rio.Component):
    shared_open_key: str = ""

    theme_variants_are_initialized: bool = False
    create_light_theme: bool = True
    create_dark_theme: bool = False

    @rio.event.on_populate
    async def _on_populate(self) -> None:
        if self.theme_variants_are_initialized:
            return

        self.theme_variants_are_initialized = True

        current_theme_is_light = self.session.theme.is_light_theme
        self.create_light_theme = current_theme_is_light
        self.create_dark_theme = not current_theme_is_light

    async def _on_radius_change(
        self,
        radius_name: str,
        event: rio.SliderChangeEvent,
    ) -> None:
        await update_and_apply_theme(
            self.session,
            {
                radius_name: event.value,
            },
        )

    async def _toggle_create_light_theme(
        self, _: rio.SwitchChangeEvent
    ) -> None:
        self.create_light_theme = not self.create_light_theme

        if not self.create_light_theme and not self.create_dark_theme:
            self.create_dark_theme = True

        if self.session.theme.is_light_theme and self.create_light_theme:
            theme_mode = "light"
        else:
            theme_mode = "dark"

        await update_and_apply_theme(
            self.session,
            {
                "mode": theme_mode,
            },
        )

    async def _toggle_create_dark_theme(self, _: rio.SwitchChangeEvent) -> None:
        self.create_dark_theme = not self.create_dark_theme

        if not self.create_light_theme and not self.create_dark_theme:
            self.create_light_theme = True

        if self.session.theme.is_light_theme and self.create_light_theme:
            theme_mode = "light"
        else:
            theme_mode = "dark"

        await update_and_apply_theme(
            self.session,
            {
                "mode": theme_mode,
            },
        )

    def build(self) -> rio.Component:
        # Prepare the radius sliders
        slider_min = 0
        slider_max = 4
        radius_sliders = rio.Grid(
            (
                rio.Text(
                    "Small",
                    justify="left",
                ),
                rio.Slider(
                    value=self.session.theme.corner_radius_small,
                    minimum=slider_min,
                    maximum=slider_max,
                    grow_x=True,
                    on_change=functools.partial(
                        self._on_radius_change,
                        "corner_radius_small",
                    ),
                ),
            ),
            (
                rio.Text(
                    "Medium",
                    justify="left",
                ),
                rio.Slider(
                    value=self.session.theme.corner_radius_medium,
                    minimum=slider_min,
                    maximum=slider_max,
                    grow_x=True,
                    on_change=functools.partial(
                        self._on_radius_change,
                        "corner_radius_medium",
                    ),
                ),
            ),
            (
                rio.Text(
                    "Large",
                    justify="left",
                ),
                rio.Slider(
                    value=self.session.theme.corner_radius_large,
                    minimum=slider_min,
                    maximum=slider_max,
                    grow_x=True,
                    on_change=functools.partial(
                        self._on_radius_change,
                        "corner_radius_large",
                    ),
                ),
            ),
            row_spacing=0.5,
        )

        # Combine everything
        return rio.ScrollContainer(
            rio.Column(
                # Main Colors
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Primary",
                    palette_slug="primary",
                    round_top=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Secondary",
                    palette_slug="secondary",
                    round_bottom=True,
                ),
                # Neutral Colors
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Background",
                    palette_slug="background",
                    margin_top=1,
                    round_top=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Neutral",
                    palette_slug="neutral",
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="HUD",
                    palette_slug="hud",
                    pick_opacity=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Disabled",
                    palette_slug="disabled",
                    round_bottom=True,
                ),
                # Semantic Colors
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Success",
                    palette_slug="success",
                    margin_top=1,
                    round_top=True,
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Warning",
                    palette_slug="warning",
                ),
                PalettePicker(
                    shared_open_key=self.bind().shared_open_key,
                    palette_nicename="Danger",
                    palette_slug="danger",
                    round_bottom=True,
                ),
                # Corner radii
                rio.Text(
                    "Corner Radii",
                    style="heading3",
                    margin_top=1,
                    margin_bottom=1,
                    justify="left",
                ),
                radius_sliders,
                # Theme Variants
                rio.Text(
                    "Variants",
                    style="heading3",
                    margin_top=1,
                    margin_bottom=1,
                    justify="left",
                ),
                rio.Grid(
                    [
                        rio.Text("Light Theme", justify="left"),
                        rio.Switch(
                            is_on=self.create_light_theme,
                            on_change=self._toggle_create_light_theme,
                            grow_x=True,
                        ),
                    ],
                    [
                        rio.Text("Dark Theme", justify="left"),
                        rio.Switch(
                            is_on=self.create_dark_theme,
                            on_change=self._toggle_create_dark_theme,
                            grow_x=True,
                        ),
                    ],
                    row_spacing=0.5,
                    column_spacing=0.5,
                ),
                # Code Sample
                rio.Text(
                    "Code",
                    style="heading3",
                    margin_top=1,
                    justify="left",
                ),
                rio.Text(
                    "Use this code to recreate the current theme in your app:",
                    justify="left",
                    margin_top=0.5,
                ),
                rio.CodeBlock(
                    get_source_for_theme(
                        self.session.theme,
                        create_theme_pair=self.create_light_theme
                        and self.create_dark_theme,
                    ),
                    margin_top=0.5,
                ),
                margin=1,
                align_y=0,
            ),
            scroll_x="never",
        )



# <additional-imports>
import rio

from .. import components as comps
from .. import data_models, theme

# </additional-imports>


# <component>
class FaqSection(rio.Component):
    """
    A responsive FAQ section displaying frequently asked questions.

    This section consists of:
    - A title and subtitle providing an introduction to the FAQ.
    - Multiple `FAQ` components for individual questions and answers.
    - Separators for visual organization.
    - Layout adjustments for desktop and mobile devices.
    """

    def _desktop_build(self) -> rio.Component:
        """
        Builds the column layout optimized for desktop devices.
        """
        return rio.Column(
            # Add Section Title
            rio.Text(
                "Frequently asked questions",
                justify="center",
                style=theme.BOLD_BIGGER_SECTION_TITEL_DESKTOP,
            ),
            # Add Section Subtitle
            rio.Text(
                "At vero eos et accusam et justo duo dolores et ea rebum.",
                justify="center",
                style=theme.DARK_TEXT_BIGGER,
            ),
            # Add FAQ items
            rio.Column(
                comps.Faq(
                    header="What is the refund policy?",
                    body="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="How can I cancel my subscription?",
                    body="Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="How does a free trial work?",
                    body="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="Can I switch plans later?",
                    body="Stet clita kasd gubergren, no sea takimata sanctus, sed diam voluptua.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="Do you offer refunds?",
                    body="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="Do you offer support?",
                    body="Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla.",
                    is_last=True,  # Indicates this is the last FAQ and adds an extra Separator when open
                ),
                margin_top=4,  # Top margin for the FAQ list
                align_x=0.5,  # Center-align the column
                min_width=60,  # Fixed width for the FAQ section
            ),
            margin_bottom=10,  # Bottom margin for the section
            margin_top=5,  # Top margin for the section
            align_x=0.5,  # Center-align the section
            min_width=80,  # Fixed width for the section
            spacing=1,  # Space between elements
            align_y=0,  # Align vertically at the top
        )

    def _mobile_build(self) -> rio.Component:
        """
        Builds the column layout optimized for mobile devices.
        """
        return rio.Column(
            # Add Section Title
            rio.Text(
                "Frequently asked questions",
                justify="center",
                overflow="wrap",
                style=theme.BOLD_SECTION_TITEL_MOBILE,
            ),
            # Add Section Subtitle
            rio.Text(
                "At vero eos et accusam et justo duo dolores et ea rebum.",
                justify="center",
                overflow="wrap",
                style=theme.DARKER_TEXT,
            ),
            # Add FAQ items
            rio.Column(
                comps.Faq(
                    header="What is the refund policy?",
                    body="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="How can I cancel my subscription?",
                    body="Stet clita kasd gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="How does a free trial work?",
                    body="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="Can I switch plans later?",
                    body="Stet clita kasd gubergren, no sea takimata sanctus, sed diam voluptua.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="Do you offer refunds?",
                    body="Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod tempor invidunt ut labore et dolore magna aliquyam.",
                ),
                rio.Separator(color=theme.NEUTRAL_COLOR_BRIGHTER),
                comps.Faq(
                    header="Do you offer support?",
                    body="Duis autem vel eum iriure dolor in hendrerit in vulputate velit esse molestie consequat, vel illum dolore eu feugiat nulla.",
                    is_last=True,  # Indicates this is the last FAQ and adds an extra Separator when open
                ),
                margin_top=4,  # Top margin for the FAQ list
            ),
            spacing=1,  # Space between elements
            align_y=0,  # Align vertically at the top
        )

    def build(self) -> rio.Component:
        """
        Build and return the page layout based on the user's device type.

        Determines whether the user is on a desktop or mobile device and builds
        the corresponding layout.
        """
        device = self.session[data_models.PageLayout].device

        if device == "desktop":
            # Build and return the desktop-optimized layout
            return self._desktop_build()

        else:
            # Build and return the mobile-optimized layout
            return self._mobile_build()


# </component>



import dataclasses
import typing as t


@dataclasses.dataclass(frozen=True)
class PageLayout:
    """
    Represents the layout configuration for a page.


    ## Attributes:

    `device`: Specifies the type of device for the page layout.
    """

    device: t.Literal["desktop", "mobile"]


@dataclasses.dataclass
class Testimonial:
    """
    Class to represent a testimonial on the website.


    ## Attributes:

    `quote`: The text of the testimonial.
    `name`: The name of the person who gave the testimonial.
    `company`: The company the person works for.
    `image`: The URL of the image of the person who gave the testimonial.
    """

    quote: str
    name: str
    company: str
    image: str


@dataclasses.dataclass
class BlogPost:
    """
    Class to represent a blog post on the website.


    ## Attributes:

    `title`: The title of the blog post.
    `description`: A short description of the blog post.
    `image`: The URL of the image of the blog post.
    `category`: The category of the blog post e.g. Nature, Technology, etc.
    `author`: The author of the blog post.
    `author_image`: The URL of the image of the author of the blog post.
    `date`: The date the blog post was published.
    `url_segment`: The URL segment of the blog post.
    """

    title: str
    description: str
    image: str
    category: str
    author: str
    author_image: str
    date: str
    url_segment: str



import json
import subprocess
import sys
import tempfile
from pathlib import Path

__all__ = ["ruff_check", "ruff_format"]


def ruff_format(source_code: str) -> str:
    # Write the source code to a temporary file
    temp_file_path = Path(tempfile.gettempdir()) / "rio test suite tempfile.py"

    temp_file_path.write_text(source_code, encoding="utf8")

    # Run ruff to format the source code in the temporary file
    ruff("format", temp_file_path)

    # Read the formatted source code
    return temp_file_path.read_text(encoding="utf8")


def ruff_check(source_code: str) -> list[str]:
    """
    Checks the given source code using `ruff`. Returns any encountered problems.
    """
    # Dump the source to a file, and implicitly define/import some stuff.
    temp_file_path = Path(tempfile.gettempdir()) / "rio test suite tempfile.py"

    temp_file_path.write_text(
        f"""
import asyncio
import pathlib
import rio

Path = pathlib.Path
app = rio.App(build=rio.Spacer)
MyAppRoot = rio.Spacer
self = child1 = child2 = rio.Spacer()

{source_code}
""",
        encoding="utf8",
    )

    # Run ruff to format the source code in the temporary file
    proc = ruff(
        "check",
        temp_file_path,
        # E402 = Import not at top of file
        #
        # F401 = Unused import
        #
        # F811 = Redefinition of a symbol. Happens if the source code already
        # includes one of our injected imports.
        "--ignore=E402,F401,F811",
        "--output-format=json",
    )

    output = json.loads(proc.stdout)
    assert isinstance(output, list), output

    # Parse the output
    return [entry["message"] for entry in output]


def ruff(*args: str | Path) -> subprocess.CompletedProcess:
    return subprocess.run(
        [sys.executable, "-m", "ruff", *map(str, args)],
        # check=True,
        capture_output=True,
        text=True,
    )



"""
Tests that different types of data formats accepted by tables are correctly
turned into the internal column format.
"""

import typing as t

import numpy as np
import pandas as pd
import polars as pl
import pytest

import rio
import rio.maybes

# Tables work with a lot of optionally supported modules. Make sure Rio is aware
# of which ones are available.
rio.maybes.initialize()


def gen_valid_data() -> dict[str, t.Iterable[t.Any]]:
    """
    Generates a dictionary of valid data for use in a `rio.Table`. The columns
    are intentionally of different types, some of which can only be used once.
    """

    return {
        "Text": ["A", "B", "C", "D", "E"],
        "Number List": [1, 2, 3, 4, 5],
        "Number Tuple": (1, 2, 3, 4, 5),
        "Number Generator": (i for i in range(1, 6)),
    }


def gen_too_short_column_data() -> dict[str, t.Iterable[t.Any]]:
    """
    Same as the function above, but one column is intentionally shorter than the
    others.
    """
    data = gen_valid_data()
    data["Short Column"] = [1, 2, 3, 4]
    return data


def gen_too_long_column_data() -> dict[str, t.Iterable[t.Any]]:
    """
    Same as the function above, but one column is intentionally longer than the
    others.
    """
    data = gen_valid_data()
    data["Long Column"] = [1, 2, 3, 4, 5, 6]
    return data


def into_all_formats(
    datagen: t.Callable[
        [],
        dict[str, t.Iterable[t.Any]],
    ],
) -> t.Iterable[tuple[t.Any, bool]]:
    """
    Given a data generator, return the same data in all input formats supported
    by `rio.Table`:

    - pandas DataFrame
    - polars DataFrame
    - numpy array
    - mapping of iterables
    - iterable of iterables

    Each result is a tuple of the data and a boolean indicating whether the
    resulting table should have headers.
    """
    # Pandas DataFrame
    yield pd.DataFrame(datagen()), True

    # Polars DataFrame
    yield pl.DataFrame(datagen()), True

    # NumPy array
    #
    # Numpy is problematic, because it cannot store different types of data in
    # the same array. It implicitly converts everything to strings, which makes
    # the later correctness checks fail.
    #
    # yield (
    #     np.column_stack(
    #         [list(column) for column in datagen().values()],
    #     ),
    #     False,
    # )

    # Mapping of iterables
    yield datagen(), True

    # Iterable of iterables
    as_df = pl.DataFrame(datagen())
    rows: list[list[t.Any]] = []

    for row in as_df.rows():
        rows.append(list(row))

    yield rows, False


def assert_columns_match_data(
    datagen: t.Callable[
        [],
        dict[str, t.Iterable[t.Any]],
    ],
    should_have_headers: bool,
    headers_are: list[str] | None,
    columns_are: list[list[t.Any]],
) -> None:
    """
    Asserts that columns in the standardized format used by `rio.Table` match
    what would be expected from the data generator.
    """
    data = datagen()

    # Do the headers match?
    if should_have_headers:
        assert headers_are is not None
        assert headers_are == list(data.keys())
    else:
        assert headers_are is None

    # Correct number of columns?
    assert len(columns_are) == len(data)

    # Column values
    for ii, (column_name, column_should) in enumerate(data.items()):
        column_should = list(column_should)
        column_is = columns_are[ii]

        assert column_is == column_should


@pytest.mark.parametrize(
    "data, should_have_headers",
    into_all_formats(gen_valid_data),
)
def test_valid_data(data: t.Any, should_have_headers: bool) -> None:
    """
    Tests that valid data is correctly columnized.
    """
    headers_are, columns_are = rio.components.table._data_to_columnar(data)

    assert_columns_match_data(
        datagen=gen_valid_data,
        should_have_headers=should_have_headers,
        headers_are=headers_are,
        columns_are=columns_are,
    )


def test_short_column() -> None:
    """
    Tests that data with a column that is too short fails as expected.
    """
    data = gen_too_short_column_data()

    # The only format supporting too short data is a mapping. All other formats
    # would immediately raise an error, before even being able to pass the data
    # to a table.
    #
    # -> No need for fancy formats, just pass in the dict directly.
    with pytest.raises(ValueError):
        rio.components.table._data_to_columnar(data)


def test_long_column() -> None:
    """
    Tests that data with a column that is too long fails as expected.
    """
    data = gen_too_long_column_data()

    # The only format supporting too short data is a mapping. All other formats
    # would immediately raise an error, before even being able to pass the data
    # to a table.
    #
    # -> No need for fancy formats, just pass in the dict
    with pytest.raises(ValueError):
        rio.components.table._data_to_columnar(data)


def test_1d_array() -> None:
    """
    Creating a table from a 1D array should fail, as tables are inherently
    two-dimensional.
    """
    with pytest.raises(ValueError):
        rio.components.table._data_to_columnar(np.array([1, 2, 3]))


def test_2d_array() -> None:
    """
    Creating a table from a 2D array should work.
    """
    headers, columns = rio.components.table._data_to_columnar(
        np.array([[1, 2, 3], [4, 5, 6]]),
    )

    assert_columns_match_data(
        datagen=lambda: {
            "0": [1, 4],
            "1": [2, 5],
            "2": [3, 6],
        },
        should_have_headers=False,
        headers_are=headers,
        columns_are=columns,
    )


def test_3d_array() -> None:
    """
    Creating a table from a 3D array should fail, as tables are inherently
    two-dimensional.
    """
    with pytest.raises(ValueError):
        rio.components.table._data_to_columnar(np.array([[[1, 2], [3, 4]]]))



from __future__ import annotations

import dataclasses
import typing as t

import rio
import rio.components.fundamental_component
import rio.debug.dev_tools.layout_display

from . import layout_explainer


class SizeControls(rio.Component):
    label: t.Literal["width", "height"]
    grow: bool
    min_value: float

    _: dataclasses.KW_ONLY

    on_grow_change: rio.EventHandler[bool] = None
    on_min_change: rio.EventHandler[float] = None

    async def _on_grow_change(
        self,
        event: rio.SwitchChangeEvent,
    ) -> None:
        self.grow = event.is_on
        await self.call_event_handler(self.on_grow_change, self.grow)

    async def _on_min_value_change(
        self,
        event: rio.NumberInputChangeEvent,
    ) -> None:
        self.min_value = event.value
        await self.call_event_handler(self.on_min_change, self.min_value)

    def build(self) -> rio.Component:
        axis_xy = "x" if self.label == "width" else "y"

        return rio.Row(
            rio.NumberInput(
                label=f"Min {self.label.capitalize()}",
                value=self.bind().min_value,
                on_change=self._on_min_value_change,
                grow_x=True,
            ),
            rio.Spacer(min_width=1, grow_x=False),
            rio.Switch(
                is_on=self.bind().grow,
                on_change=lambda event: self._on_grow_change,
            ),
            rio.Text(f"Grow {axis_xy.capitalize()}"),
            spacing=0.5,
        )


class AlignmentControls(rio.Component):
    label: str
    value: float | None

    _: dataclasses.KW_ONLY

    on_change: rio.EventHandler[float | None] = None

    async def _on_switch_change(self, event: rio.SwitchChangeEvent) -> None:
        if event.is_on:
            self.value = 0.5
        else:
            self.value = None

        await self.call_event_handler(self.on_change, self.value)

    async def _on_slider_change(self, event: rio.SliderChangeEvent) -> None:
        self.value = event.value
        await self.call_event_handler(self.on_change, self.value)

    def build(self) -> rio.Component:
        # If enabled, add a slider to control the alignment value
        if self.value is None:
            slider = None
        else:
            slider = rio.Slider(
                value=self.value,
                minimum=0,
                maximum=1,
                show_values=True,
                on_change=self._on_slider_change,
                margin_top=1,
                margin_bottom=0.5,
            )

        # Allow enabling / disabling alignment
        return rio.Grid(
            [
                rio.Switch(
                    is_on=self.value is not None,
                    on_change=self._on_switch_change,
                ),
                rio.Text(
                    self.label,
                    align_x=0,
                    grow_x=True,
                ),
            ],
            rio.Switcher(slider),
            row_spacing=0.5,
            column_spacing=0.5,
        )


class HelpAnchor(rio.Component):
    content: str

    def build(self) -> rio.Component:
        return rio.Tooltip(
            # The rectangle increases the area the user can hover over
            anchor=rio.Rectangle(
                content=rio.Row(
                    rio.Icon("material/help", fill="secondary"),
                    rio.Text(
                        "Help",
                        style=rio.TextStyle(
                            fill=self.session.theme.secondary_color
                        ),
                    ),
                    spacing=0.5,
                    margin_y=0.5,
                    align_x=0,
                ),
                fill=rio.Color.TRANSPARENT,
            ),
            tip=rio.Markdown(
                self.content,
                min_width=25,
            ),
            position="left",
            gap=2,
            align_x=0,
        )


class ActionAnchor(rio.Component):
    icon: str
    partial_name: str
    actions: list[str]

    def build(self) -> rio.Component:
        # Prepare the anchor
        if len(self.actions) == 0:
            color = self.session.theme.disabled_color
        else:
            color = self.session.theme.secondary_color

        anchor = rio.Row(
            rio.Icon(self.icon, fill=color),
            rio.Text(
                f"How to {self.partial_name}",
                style=rio.TextStyle(fill=color),
            ),
            spacing=0.3,
            align_x=0.5,
            margin=0.5,
        )

        # If there aren't any options there is no need to go any further
        if not self.actions:
            return anchor

        # Prepare the markdown content
        if len(self.actions) == 1:
            markdown_source = self.actions[0]
        else:
            markdown_source = "- " + "\n- ".join(self.actions)

        # Build the UI
        return rio.Tooltip(
            anchor=anchor,
            tip=rio.Markdown(
                markdown_source,
                min_width=25,
            ),
            position="left",
            gap=-2,
        )


class LayoutSubpage(rio.Component):
    component_id: int

    _: dataclasses.KW_ONLY

    _layout_explainer: layout_explainer.LayoutExplainer | None = None
    _explanation_is_expanded: bool = True

    def get_target_component(self) -> rio.Component:
        return self.session._weak_components_by_id[self.component_id]

    @rio.event.on_populate
    async def _update_explanations(self) -> None:
        # Get human-readable layout explanations
        target = self.get_target_component()

        try:
            self._layout_explainer = (
                await layout_explainer.LayoutExplainer.create_new(
                    self.session,
                    target,
                )
            )
        # This is raised if the component cannot be found client-side, e.g. due
        # to network lag
        except KeyError:
            self._layout_explainer = None

    async def _update_target_attribute(self, name: str, value: t.Any) -> None:
        """
        Updates an attribute of the target component.

        This is more complicated than just setting the attribute, because Rio
        must be notified of the change in order to update the display.
        """
        target = self.get_target_component()

        # Assign the new value to the Python instance
        setattr(target, name, value)

        # Components will automatically mark themselves as dirty, but won't
        # trigger a resync. Do that now.
        await self.session._refresh()

        # Update the explanations
        await self._update_explanations()

    def _build_explanations(self) -> rio.Component:
        if self._layout_explainer is None:
            return rio.Text(
                "No explanations available",
                style="dim",
                justify="left",
                margin_y=0.5,
            )

        content = rio.Column(
            spacing=0.5,
            # Push away the silly rounded corner created by the revealer. It
            # looks silly when cutting off warnings.
            margin_bottom=0.5,
        )

        # Explain the width
        content.add(
            rio.Markdown(
                f"**Width**: {self._layout_explainer.width_explanation}",
            )
        )

        content.add(
            rio.Row(
                ActionAnchor(
                    "close-fullscreen",
                    "shrink width",
                    self._layout_explainer.decrease_width,
                ),
                ActionAnchor(
                    "open-in-full",
                    "grow width",
                    self._layout_explainer.increase_width,
                ),
            )
        )

        # Explain the height
        content.add(
            rio.Markdown(
                f"**Height**: {self._layout_explainer.height_explanation}",
            )
        )

        content.add(
            rio.Row(
                ActionAnchor(
                    "close-fullscreen",
                    "shrink height",
                    self._layout_explainer.decrease_height,
                ),
                ActionAnchor(
                    "open-in-full",
                    "grow height",
                    self._layout_explainer.increase_height,
                ),
            )
        )

        # Add warnings
        for warning in self._layout_explainer.warnings:
            content.add(
                rio.Row(
                    rio.Rectangle(
                        fill=self.session.theme.warning_color,
                        min_width=0.3,
                    ),
                    rio.Markdown(
                        warning,
                        grow_x=True,
                    ),
                    spacing=0.5,
                    margin_top=0.5,
                )
            )

        return content

    def _build_margin_controls(self) -> rio.Component:
        result = rio.Grid(
            row_spacing=0.3,
            column_spacing=0.3,
        )

        # Add a visual stand-in for the target component
        target = self.get_target_component()
        result.add(
            rio.Card(
                content=rio.Text(
                    type(target).__name__,
                    justify="center",
                ),
                color="secondary",
                corner_radius=self.session.theme.corner_radius_small,
            ),
            1,
            1,
        )

        # And controls for the margin
        result.add(
            rio.NumberInput(
                label="Left",
                value=target._effective_margin_left_,
                minimum=0,
                on_change=lambda event: self._update_target_attribute(
                    "margin_left", event.value
                ),
            ),
            1,
            0,
        )

        result.add(
            rio.NumberInput(
                label="Top",
                value=target._effective_margin_top_,
                minimum=0,
                on_change=lambda event: self._update_target_attribute(
                    "margin_top", event.value
                ),
            ),
            0,
            1,
        )

        result.add(
            rio.NumberInput(
                label="Right",
                value=target._effective_margin_right_,
                minimum=0,
                on_change=lambda event: self._update_target_attribute(
                    "margin_right", event.value
                ),
            ),
            1,
            2,
        )

        result.add(
            rio.NumberInput(
                label="Bottom",
                value=target._effective_margin_bottom_,
                minimum=0,
                on_change=lambda event: self._update_target_attribute(
                    "margin_bottom", event.value
                ),
            ),
            2,
            1,
        )

        result.add(
            HelpAnchor(
                """
Margin adds empty space around the component. You can either specify each side
separately, or use one of the shortcuts `margin`, `margin_x`, `margin_y`.
                """,
            ),
            3,
            0,
            width=3,
        )

        return result

    def _build_size_controls(self) -> rio.Component:
        target_component = self.get_target_component()

        return rio.Column(
            SizeControls(
                label="width",
                grow=target_component.grow_x,
                min_value=target_component.min_width,
                on_grow_change=lambda value: self._update_target_attribute(
                    "grow_x", value
                ),
                on_min_change=lambda value: self._update_target_attribute(
                    "min_width", value
                ),
            ),
            SizeControls(
                label="height",
                grow=target_component.grow_y,
                min_value=target_component.min_height,
                on_grow_change=lambda value: self._update_target_attribute(
                    "grow_y", value
                ),
                on_min_change=lambda value: self._update_target_attribute(
                    "min_height", value
                ),
            ),
            HelpAnchor(
                """
By default, components take up as little space as necessary. You can set a
custom `min_width` and `min_height` to make them take up more space.

Components with `grow_x` or `grow_y` take priority when too much space is
available. This is only relevant in components that have multiple children, such
as `Row` and `Column`.

Above all, **components always take up all superfluous space from their
parent**. If you don't want for that to happen, set an alignment.
                """,
            ),
            spacing=0.5,
        )

    def _build_alignment_controls(self) -> rio.Component:
        target = self.get_target_component()

        return rio.Column(
            AlignmentControls(
                label="Align X",
                value=target.align_x,
                on_change=lambda value: self._update_target_attribute(
                    "align_x", value
                ),
            ),
            AlignmentControls(
                label="Align Y",
                value=target.align_y,
                on_change=lambda value: self._update_target_attribute(
                    "align_y", value
                ),
            ),
            HelpAnchor(
                """
Unaligned components take up the full space assigned by their parent. Aligned
components only take up as little space as necessary to display their content.

Components with an alignment of `0` are left/top-aligned, `0.5` are centered,
and `1` are right/bottom-aligned.
                    """
            ),
            margin_top=0.3,
            spacing=0.5,
        )

    def build(self) -> rio.Component:
        return rio.Column(
            rio.debug.dev_tools.layout_display.LayoutDisplay(
                component_id=self.bind().component_id,
                max_requested_height=20,
                min_height=20,
                on_component_change=lambda _: self._update_explanations(),
                on_layout_change=self._update_explanations,
            ),
            rio.Revealer(
                header="Explanation",
                content=self._build_explanations(),
                header_style="heading3",
                is_open=self.bind()._explanation_is_expanded,
            ),
            rio.Revealer(
                header="Margin",
                content=self._build_margin_controls(),
                header_style="heading3",
            ),
            rio.Revealer(
                header="Size",
                content=self._build_size_controls(),
                header_style="heading3",
            ),
            rio.Revealer(
                header="Alignment",
                content=self._build_alignment_controls(),
                header_style="heading3",
            ),
            rio.Spacer(),
            rio.Link(
                "Layouting Quickstart",
                icon="material/library_books",
                target_url="https://rio.dev/docs/howto/layout-guide?s=w1q",
                open_in_new_tab=True,
                margin_top=1,
                align_x=0,
            ),
            spacing=0.5,
        )



import typing as t

import pytest

import rio
from tests.utils.layouting import verify_layout


@pytest.mark.parametrize(
    "scroll_x,scroll_y",
    [
        ("never", "auto"),
        ("auto", "never"),
        ("auto", "auto"),
    ],
)
async def test_scrolling(
    scroll_x: t.Literal["never", "always", "auto"],
    scroll_y: t.Literal["never", "always", "auto"],
) -> None:
    await verify_layout(
        lambda: rio.ScrollContainer(
            rio.Text("hi", min_width=30, min_height=30),
            scroll_x=scroll_x,
            scroll_y=scroll_y,
            min_width=20,
            min_height=20,
            align_x=0.5,
            align_y=0.5,
        )
    )



import gc
import random
import asyncio
import json
import rio
from rio.theme import _derive_color, _make_semantic_palette
import rio.components.class_container
import rio.data_models
import rio.debug
import rio.debug.dev_tools
import rio.debug.dev_tools.dev_tools_connector
import rio.debug.dev_tools.icons_page
import rio.debug.dev_tools.layout_display
import rio.debug.layouter
import typing as t


class ThemeSample(rio.Component):
    theme: rio.Theme

    async def _apply_theme_worker(self) -> None:
        # Wait for a bit to ensure the client has time to update the HTML
        await asyncio.sleep(0.2)

        # Get all variables to apply
        light_or_dark = "light" if self.theme.is_light_theme else "dark"
        theme_vars = self.session._calculate_theme_css_values(self.theme)

        # Apply the theme
        await self.session._evaluate_javascript(
            f"""
let themeVars = {json.dumps(theme_vars)};

let elem = globalThis.componentsById[{self._id}].element;

for (let key in themeVars) {{
    elem.style.setProperty(
        key,
        themeVars[key],
    );
}}

// Set the theme variant
document.documentElement.setAttribute(
    "data-theme",
    {json.dumps(light_or_dark)}
);

            """
        )

    def build(self) -> rio.Component:
        # Apply the theme
        asyncio.create_task(
            self._apply_theme_worker(),
        )

        # Build the content
        content = rio.Row(
            # Main Content
            rio.Column(
                rio.Text(
                    "Lorem Ipsum",
                    style="heading1",
                ),
                rio.Markdown(
                    """
Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam nec
ultricies elit. Nullam nec ultricies elit. Nullam nec ultricies elit.
                    """,
                ),
                rio.Image(
                    rio.URL("https://picsum.photos/300/200"),
                    min_width=26,
                    min_height=14,
                    corner_radius=self.theme.corner_radius_medium,
                ),
                rio.Markdown(
                    """
Lorem markdownum auctor fiant aversa, scopulos bos ima ruptosque turbam, fertur
flexile amictu temptaretque *superum temptarunt* volatu ad veros. Fatebor
Aesacos.

```js
rom_root(record);
direct = subdirectory(symbolic_restore_t, url);
```

Coniuge et tamen et omnibus ille concordant mitissima accedere fratrem
auditurum, spes petunt ore insigne mille. Te explorant ardor, dies inde eloquio
rostrum vestemque et frustra purum. Cynthia quoniam dextra aequoris, emissi [in
                    """,
                ),
                grow_x=True,
            ),
            # (Fake) Drawer
            rio.ThemeContextSwitcher(
                content=rio.Rectangle(
                    content=rio.Grid(
                        rio.Text(
                            "Drawer Heading",
                            style="heading2",
                        ),
                        rio.TextInput(
                            text="Foobar",
                            style="rounded",
                        ),
                        rio.TextInput(
                            text="Foobar",
                            style="rounded",
                            is_sensitive=False,
                        ),
                        [
                            rio.Text("Switch"),
                            rio.Switch(),
                        ],
                        [
                            rio.Button("Major", style="major"),
                            rio.Button("Minor", style="minor"),
                        ],
                        [
                            rio.Button("Colored Text", style="colored-text"),
                            rio.Button("Plain Text", style="plain-text"),
                        ],
                        rio.Switch(is_on=True),
                        rio.Switch(is_on=False),
                        rio.Switch(is_on=True, is_sensitive=False),
                        rio.Switch(is_on=False, is_sensitive=False),
                        rio.Button("Enabled"),
                        rio.Button("Disabled", is_sensitive=False),
                        rio.Spacer(),
                        rio.Separator(),
                        rio.Row(
                            rio.Icon("material/castle"),
                            rio.Text(
                                "Castles are Cool",
                                grow_x=True,
                            ),
                            spacing=0.5,
                        ),
                        row_spacing=1,
                        column_spacing=1,
                        margin=2,
                    ),
                    fill=self.theme.neutral_color,
                    corner_radius=(
                        self.theme.corner_radius_large,
                        0,
                        0,
                        self.theme.corner_radius_large,
                    ),
                    shadow_color=self.theme.shadow_color,
                    shadow_radius=1.3,
                ),
                color="neutral",
            ),
            spacing=1,
        )

        # Wrap up the content
        #
        # A rectangle acts as a fake background
        return rio.Rectangle(
            # A Theme context switcher ensures that the theme variables are
            # locally exposed.
            content=rio.ThemeContextSwitcher(
                content=content,
                color="background",
            ),
            fill=self.theme.background_color,
        )


class ThemeViewer(rio.Component):
    def build(self) -> rio.Component:
        # Decide on the colors to use
        theme_colors: list[rio.Color] = [
            self.session.theme.primary_color,
            rio.Color.RED,
            rio.Color.GREEN,
            rio.Color.YELLOW,
        ]

        # Create the themes
        light_viewers: list[ThemeSample] = []
        dark_viewers: list[ThemeSample] = []

        for col in theme_colors:
            light_theme = rio.Theme.from_colors(
                primary_color=col,
                mode="light",
            )

            dark_theme = rio.Theme.from_colors(
                primary_color=col,
                mode="dark",
            )

            light_viewers.append(ThemeSample(theme=light_theme))
            dark_viewers.append(ThemeSample(theme=dark_theme))

        # Populate
        return rio.Grid(
            light_viewers,
            dark_viewers,
            row_spacing=0.4,
            column_spacing=0.4,
        )


app = rio.App(
    build=ThemeViewer,
    theme=(
        rio.Theme.from_colors(mode="light"),
        rio.Theme.from_colors(mode="dark"),
    ),
    default_attachments=[],
)



class NoFileSelectedError(Exception):
    """
    Raised when the user doesn't select a file.

    This exception is raised when the user is prompted to select a file with a
    file picker and they close the file picker without selecting a file. This is
    a perfectly valid operation on the user's end - make sure you handle it.
    """

    pass


class AssetError(Exception):
    """
    Raised when an error occurs related to assets.

    Signifies that some operation related to assets has failed. E.g. trying to
    access a nonexistent asset.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        """
        A human-readable message describing the problem.

        This message contains additional information about the error to inform
        the developer.
        """
        return self.args[0]


class NavigationFailed(Exception):
    """
    Raised when navigation to a page fails.

    This exception is raised when attempting to navigate to a page, but the
    navigation fails for some reason. This could happen, for example, because a
    page guard throws an exception.

    Note that navigating to a nonexistent pages is not an error, as `PageViews`
    will simply display their fallback in that case. Thus this exception will
    not be raised in that case.
    """


class ClipboardError(Exception):
    """
    Raised for errors related to clipboard operations.

    This exception is raised when an error occurs while trying to interact with
    the clipboard. For example when trying to read from the clipboard and the
    browser refuses due to privacy restrictions.
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        """
        Returns the error message as a string.

        This message contains additional information about the error in a form
        amenable to being displayed to humans.
        """
        return self.args[0]


class InvalidProjectConfigError(Exception):
    """
    Raised when the project configuration is invalid.

    This exception is raised when a function cannot continue due to a problem
    with the project configuration. For example, the project's `rio.toml` could
    be invalid TOML, or not exist at all.

    ## Metadata

    `public`: False
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)

    @property
    def message(self) -> str:
        """
        Returns the error message as a string.
        """
        return self.args[0]



from __future__ import annotations

import asyncio
import dataclasses
import logging
import secrets
import typing as t
from base64 import urlsafe_b64encode
from datetime import timedelta
from hashlib import sha256

import fastapi
import timer_dict
from authlib.integrations.httpx_client import AsyncOAuth2Client

import rio

# Define constants for OAuth2 client configuration
CLIENT_ID = "..."
CLIENT_SECRET = "..."
SCOPE = "user:email"
REDIRECT_URI = "http://localhost:8001/callback"


@dataclasses.dataclass
class OAuth:
    _extension: OAuthExtension
    _session: rio.Session

    _state: t.Literal["logged-out", "logging-in", "logged-in"]

    @property
    def is_logged_in(self) -> bool:
        """
        Returns whether the client is currently authenticated.

        Returns `True` if the user has successfully been authenticated and
        `False` otherwise. This property is read-only. To log in, use the
        `login` method.

        Note that this will return `False` even if a login is currently in
        progress. You can use `is_logging_in` to check if that's the case.
        """
        return self._state == "logged-in"

    @property
    def is_logging_in(self) -> bool:
        """
        Returns whether the client is currently logging in.

        Returns `True` if the client is currently in the process of logging in
        and `False` otherwise.
        """
        return self._state == "logging-in"

    async def login(self) -> None:
        # Can't log in twice
        if self._state == "logging-in":
            raise RuntimeError("Already logging in")

        if self._state == "logged-in":
            raise RuntimeError("Already logged in")

        # Logging in now
        self._state = "logging-in"

        # Generate a code verifier and challenge for PKCE
        code_verifier = secrets.token_urlsafe(100)
        code_challenge = sha256(code_verifier.encode("utf-8")).digest()
        code_challenge = (
            urlsafe_b64encode(code_challenge).decode("utf-8").rstrip("=")
        )

        # Create an authorization URL and state using the OAuth2 client
        async with AsyncOAuth2Client(CLIENT_ID, CLIENT_SECRET) as oauth:
            auth_url, state = oauth.create_authorization_url(
                "https://github.com/login/oauth/authorize",
                code_challenge=code_challenge,
                code_challenge_method="S256",
                redirect_uri=REDIRECT_URI,
                scope=SCOPE,  # Add scope here
            )

            # Have the user log in by redirecting to the authorization URL
            self._session.navigate_to(auth_url)

            # Wait for the callback to be received
            future = asyncio.get_event_loop().create_future()
            self._extension._pending_requests[state] = (self, future)
            code = await future

            # Create a new OAuth2 client and fetch the access token
            token = await oauth.fetch_token(
                "https://github.com/login/oauth/access_token",
                code=code,
                code_verifier=code_verifier,
                grant_type="authorization_code",
                redirect_uri=REDIRECT_URI,
            )

            # Update state
            self._state = "logged-in"

            # Fetch the user's email from GitHub
            response = await oauth.get("https://api.github.com/user")
            user_data = response.json()
            print(f"Logged in as {user_data}")

    async def logout(self) -> None:
        raise NotImplementedError("TODO")


class OAuthExtension(rio.Extension):
    """
    Easily integrate OAuth2 with Rio.

    This extension provides a simple way to authenticate users using OAuth2
    in Rio applications. It handles the OAuth2 flow, including generating the
    authorization URL, handling the callback, and fetching the access token.

    ```python
    TODO: Example usage
    ```
    """

    def __init__(self) -> None:
        # This stores currently pending authentication requests. The key is
        # the state, and the value is a tuple containing
        #
        # - The OAuth object that is waiting for the callback
        # - The future that will be set when the callback is received
        #
        # The future's parameters are the code and state received from the
        # callback.
        self._pending_requests: timer_dict.TimerDict[
            str, tuple[OAuth, asyncio.Future[str]]
        ] = timer_dict.TimerDict(
            default_duration=timedelta(minutes=5),
        )

    async def _callback_route(
        self,
        request: fastapi.Request,
    ) -> fastapi.responses.Response:
        # TODO: Come up with better responses

        # Get the code & state from the query parameters
        try:
            code = request.query_params["code"]
        except KeyError:
            logging.error(
                f"Received invalid OAuth callback. `{request.url}` did not contain a `code` query parameter."
            )
            return fastapi.responses.Response(
                content='TODO: Missing "code" query parameter',
                status_code=400,
            )

        try:
            state = request.query_params["state"]
        except KeyError:
            logging.error(
                f"Received invalid OAuth callback. `{request.url}` did not contain a `state` query parameter."
            )
            return fastapi.responses.Response(
                content='TODO: Missing "state" query parameter',
                status_code=400,
            )

        # Is an OAuth object waiting for this callback?
        try:
            oauth, future = self._pending_requests.pop(state)
        except KeyError:
            logging.error(
                f"Received invalid OAuth callback for state `{state}`. There is no pending authentication request with that state."
            )
            return fastapi.responses.Response(
                content="TODO: Invalid state",
                status_code=400,
            )

        # Update that object
        future.set_result(code)

        # Done
        return fastapi.responses.Response(
            content="TODO: You can close this tab now",
            status_code=200,
        )

    @rio.extension_event.on_as_fastapi
    def on_as_fastapi(self, event: rio.ExtensionAsFastapiEvent) -> None:
        # Add a route to handle the OAuth2 callback
        event.fastapi_app.add_api_route(
            "/callback",
            self._callback_route,
            methods=["GET"],
        )

    @rio.extension_event.on_session_start
    def on_session_start(self, event: rio.ExtensionSessionStartEvent) -> None:
        # Attach an object to the session for handling authentication
        event.session.attach(
            OAuth(
                _extension=self,
                _session=event.session,
                _state="logged-out",
            )
        )


# Define a custom component class for the root of the application
class MyRoot(rio.Component):
    logged_in_email: str | None = None

    # Asynchronous method to handle the login process
    async def on_login(self) -> None:
        # Get the OAuth object from the session
        oauth = self.session[OAuth]

        # Authenticate
        await oauth.login()

    # Method to build the UI component
    def build(self) -> rio.Component:
        if self.logged_in_email is None:
            # Display a login button if the user is not logged in
            return rio.Button(
                "Login with GitHub",
                icon="brand/github",
                on_press=self.on_login,
                align_x=0.5,
                align_y=0.5,
            )
        # Display a welcome message if the user is logged in
        return rio.Text(
            f"Hello, {self.logged_in_email}!",
            justify="center",
            align_x=0.5,
            align_y=0.5,
        )


# Create a Rio application with the custom root component
app = rio.App(
    build=MyRoot,
)


app._add_extension(OAuthExtension())



from __future__ import annotations

# <additional-imports>
import rio

# </additional-imports>


# <component>
@rio.page(
    name="Blog",
    url_segment="blog",
)
class BlogPage(rio.Component):
    """
    This page will be used as the root component for blog pages. This means,
    that it will always be visible, regardless of which sub page is currently
    active.

    Similar to the root component the blog page contains a `rio.PageView`. Page
    views don't have any appearance on their own, but they are used to display
    the content of the currently active page.
    """

    def build(self) -> rio.Component:
        return rio.PageView()


# </component>



from __future__ import annotations

import dataclasses
import random
import typing as t
from datetime import date, datetime

import imy.docstrings

import rio

from .. import components as comps
from .component import Component

__all__ = [
    "DateInput",
    "DateConfirmEvent",
]


@t.final
@imy.docstrings.mark_constructor_as_private
@dataclasses.dataclass
class DateConfirmEvent:
    """
    Holds information regarding a date confirm event.

    This is a simple dataclass that stores useful information for when the user
    confirms the date in `DateInput`. You'll typically receive this as argument
    in `on_confirm` events.

    ## Attributes

    `value`: The newly selected date.
    """

    value: date


@t.final
class DateInput(Component):
    """
    Allows the user to pick a date from a calendar.

    DateInputs are similar in appearance to `TextInput` and `NumberInput`, but
    allow the user to pick a date from a calendar or enter a date, rather than
    text or number. When pressed, a calendar will pop-up, allowing the user to
    select a date.

    This makes for a compact component, which still allows the user to visually
    select a date.

    For a larger component which permanently displays a calendar, consider using
    `rio.Calendar`.


    ## Attributes

    `value`: The currently selected date.

    `label`: A short text to display next to the input field.

    `accessibility_label`: A short text describing this component for screen
        readers. If omitted, the `label` text is used.

    `style`: Changes the visual appearance of the date input.

    `on_change`: Triggered whenever the user selects a new date.

    `on_confirm`: Triggered when the user explicitly confirms their input,
            such as by pressing the "Enter" key. You can use this to trigger
            followup actions, such as logging in or submitting a form.


    ## Examples

    Here's a simple example that allows the user to select a data and displays
    it back to them:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def build(self) -> rio.Component:
            return rio.Column(
                rio.DateInput(
                    # In order to retrieve a value from the component, we'll
                    # use an attribute binding. This way our own value will
                    # be updated whenever the user changes the value.
                    value=self.bind().value,
                    label="Pick a Date",
                ),
                rio.Text(f"You've selected: {self.value}"),
            )
    ```

    Alternatively you can also attach an event handler to react to changes. This
    is a little more verbose, but allows you to run arbitrary code when the user
    picks a new date:

    ```python
    from datetime import date

    class MyComponent(rio.Component):
        value: date = date.today()

        def on_value_change(self, event: rio.DateChangeEvent):
            # This function will be called whenever the input's value
            # changes. We'll display the new value in addition to updating
            # our own attribute.
            self.value = event.value
            print(f"You've selected: {self.value}")

        def build(self) -> rio.Component:
            return rio.DateInput(
                value=self.value,
                label="Pick a Date",
                on_change=self.on_value_change,
            )
    ```


    ## Metadata

    `experimental`: True
    """

    value: date

    _: dataclasses.KW_ONLY

    label: str = ""
    accessibility_label: str = ""
    style: t.Literal["underlined", "rounded", "pill"] = "underlined"

    on_change: rio.EventHandler[rio.DateChangeEvent] = None
    on_confirm: rio.EventHandler[DateConfirmEvent] = None

    # Hide internal attributes from the type checker
    if not t.TYPE_CHECKING:
        _is_open: bool = False

    def _try_set_value(self, raw_value: str) -> bool:
        """
        Parse the given string and update the component's value accordingly.
        Returns `True` if the value was successfully updated, `False` otherwise.
        """
        # try to parse the date string
        try:
            # Maybe we can do some more advanced parsing here, but OK for now
            date_value = datetime.strptime(
                raw_value, self.session._date_format_string
            ).date()

            # Update the value
            self.value = date_value
            return True

        except ValueError:
            # Force the old value to stay
            self.value = self.value
            return False

    async def _on_value_change(self, event: rio.DateChangeEvent) -> None:
        # Close the date picker
        self._is_open = False

        # Chain the event handler
        await self.call_event_handler(self.on_change, event)

    async def _on_confirm(self, ev: rio.TextInputConfirmEvent) -> None:
        # Close the date picker
        self._is_open = False

        was_updated = self._try_set_value(ev.text)

        # Chain the event handler
        if was_updated:
            await self.call_event_handler(
                self.on_change,
                rio.DateChangeEvent(self.value),
            )

            await self.call_event_handler(
                self.on_confirm,
                DateConfirmEvent(self.value),
            )

    def _on_gain_focus(self, _: rio.TextInputFocusEvent) -> None:
        self._is_open = True

    def _on_close(self) -> None:
        self._is_open = False

    def build(self) -> rio.Component:
        return rio.Popup(
            # Display the input field
            anchor=rio.Stack(
                rio.TextInput(
                    label=self.label,
                    text=self.value.strftime(self.session._date_format_string),
                    on_gain_focus=self._on_gain_focus,
                    on_confirm=self._on_confirm,
                    style=self.style,
                    min_width=11,
                ),
                rio.Icon(
                    "material/calendar_today:fill",
                    fill="dim",
                    min_width=1.5,
                    min_height=1.5,
                    margin_right=0.5,
                    align_y=0.5,
                    align_x=1,
                ),
            ),
            # Display a calendar so the user can pick a date
            content=rio.Column(
                comps.Calendar(
                    value=self.bind().value,
                    on_change=self._on_value_change,
                    # Use a fresh calendar each time, so it displays the
                    # currently selected value each time the popup opens
                    key=str(random.random()),
                ),
                rio.Button(
                    "Cancel",
                    style="colored-text",
                    on_press=self._on_close,
                ),
                margin=1,
            ),
            color="neutral",
            position="auto",
            alignment=0.5,
            is_open=self._is_open,
            user_closable=True,
            modal=False,
        )



import rio
import typing as t
from pathlib import Path


class LoginPage(rio.Component):
    text_value: str = ""

    def build(self) -> rio.Component:
        return rio.TextInput(
            self.bind().text_value,
            on_confirm=lambda e: None,
            align_x=0.5,
            align_y=0.5,
            min_width=30,
        )


app = rio.App(
    build=LoginPage,
)



import asyncio
import socket
import typing as t

import revel
import uvicorn
import uvicorn.lifespan.on
from starlette.types import Receive, Scope, Send

import rio
import rio.app_server.fastapi_server
import rio.cli

from ... import nice_traceback, utils
from . import run_models


class UvicornWorker:
    def __init__(
        self,
        *,
        push_event: t.Callable[[run_models.Event], None],
        app_server: rio.app_server.fastapi_server.FastapiServer,
        socket: socket.socket,
        quiet: bool,
        debug_mode: bool,
        run_in_window: bool,
        on_server_is_ready_or_failed: asyncio.Future[None],
        base_url: rio.URL | None,
    ) -> None:
        self.push_event = push_event
        self.socket = socket
        self.quiet = quiet
        self.debug_mode = debug_mode
        self.run_in_window = run_in_window
        self.on_server_is_ready_or_failed = on_server_is_ready_or_failed
        self.base_url = base_url

        # The app server used to host the app.
        #
        # While already provided in the constructor, this needs some values to
        # be overridden. `replace_app` already handles this, so delegate to that
        # function.
        self.app_server = app_server
        self.replace_app(app_server)

    async def run(self) -> None:
        rio.cli._logger.debug("Uvicorn worker is starting")

        # Instead of using the ASGI app directly, create a transparent shim that
        # redirect's to the worker's currently stored app server. This allows
        # replacing the app server at will because the shim always remains the
        # same.
        #
        # ASGI is a bitch about function signatures. This function cannot be a
        # simple method, because the added `self` parameter seems to confused
        # whoever the caller is. Hence the nested function.
        async def _asgi_shim(
            scope: Scope,
            receive: Receive,
            send: Send,
        ) -> None:
            assert self.app_server is not None
            await self.app_server(scope, receive, send)

        # Set up a uvicorn server, but don't start it yet
        config = uvicorn.Config(
            app=_asgi_shim,
            log_config=None,  # Prevent uvicorn from configuring global logging
            log_level="error" if self.quiet else "info",
            timeout_graceful_shutdown=1,  # Without a timeout the server sometimes deadlocks
        )
        self._uvicorn_server = uvicorn.Server(config)

        # FIXME: Uvicorn wishes to set up the event loop. The current code
        # doesn't let it do that, since asyncio is already running.
        #
        # self._uvicorn_server.config.setup_event_loop()

        # Uvicorn can't handle `CancelledError` properly, so make sure it never
        # sees one. Uvicorn is wrapped in a task, which can finish peacefully.
        serve_task = asyncio.create_task(
            self._uvicorn_server.serve(
                sockets=[self.socket],
            ),
            name="uvicorn serve",
        )

        # Uvicorn doesn't handle CancelledError properly, which results in ugly
        # output in the console. This monkeypatch suppresses that.
        original_receive = uvicorn.lifespan.on.LifespanOn.receive

        async def patched_receive(self) -> t.Any:
            try:
                return await original_receive(self)
            except asyncio.CancelledError:
                return {
                    "type": "lifespan.shutdown",
                }

        uvicorn.lifespan.on.LifespanOn.receive = patched_receive

        # Run the server
        try:
            await asyncio.shield(serve_task)
        except asyncio.CancelledError:
            pass
        except Exception as err:
            rio.cli._logger.exception(f"Uvicorn has crashed")

            revel.error(f"Uvicorn has crashed:")
            nice_traceback.print_exception(err)
            self.push_event(run_models.StopRequested())
        finally:
            rio.cli._logger.debug("Requesting uvicorn to exit")

            self._uvicorn_server.should_exit = True

            # Make sure not to return before the task has returned, but also
            # avoid receiving any exceptions _again_
            try:
                await serve_task
            except:
                pass
            finally:
                rio.cli._logger.debug("Uvicorn serve task has ended")

    def replace_app(
        self,
        app_server: rio.app_server.fastapi_server.FastapiServer,
    ) -> None:
        """
        Replace the app currently running in the server with a new one. The
        worker must already be running for this to work.
        """
        assert (
            app_server.internal_on_app_start is None
        ), app_server.internal_on_app_start

        # Store the new app
        self.app_server = app_server

        self.app_server.debug_mode = self.debug_mode
        self.app_server.running_in_window = self.run_in_window
        self.app_server.internal_on_app_start = (
            lambda: self.on_server_is_ready_or_failed.set_result(None)
        )

        if self.base_url is None:
            self.app_server.base_url = None
        else:
            self.app_server.base_url = utils.normalize_url(self.base_url)

        # There is no need to inject the new app or server anywhere. Since
        # uvicorn was fed a shim function instead of the app directly, any
        # requests will automatically be redirected to the new server instance.



# <additional-imports>
from __future__ import annotations

from datetime import datetime, timedelta, timezone

import rio

from . import data_models, persistence

# </additional-imports>


# <additional-code>
async def on_app_start(app: rio.App) -> None:
    # Create a persistence instance. This class hides the gritty details of
    # database interaction from the app.
    pers = persistence.Persistence()

    # Now attach it to the session. This way, the persistence instance is
    # available to all components using `self.session[persistence.Persistence]`
    app.default_attachments.append(pers)


async def on_session_start(rio_session: rio.Session) -> None:
    # A new user has just connected. Check if they have a valid auth token.
    #
    # Any classes inheriting from `rio.UserSettings` will be automatically
    # stored on the client's device when attached to the session. Thus, by
    # retrieving the value here, we can check if the user has a valid auth token
    # stored.
    user_settings = rio_session[data_models.UserSettings]

    # Get the persistence instance
    pers = rio_session[persistence.Persistence]

    # Try to find a session with the given auth token
    try:
        user_session = await pers.get_session_by_auth_token(
            user_settings.auth_token,
        )

    # None was found - this auth token is invalid
    except KeyError:
        pass

    # A session was found. Welcome back!
    else:
        # Make sure the session is still valid
        if user_session.valid_until > datetime.now(tz=timezone.utc):
            # Attach the session. This way any component that wishes to access
            # information about the user can do so.
            rio_session.attach(user_session)

            # For a user to be considered logged in, a `UserInfo` also needs to
            # be attached.
            userinfo = await pers.get_user_by_id(user_session.user_id)
            rio_session.attach(userinfo)

            # Since this session has only just been used, let's extend its
            # duration. This way users don't get logged out as long as they keep
            # using the app.
            await pers.update_session_duration(
                user_session,
                new_valid_until=datetime.now(tz=timezone.utc)
                + timedelta(days=7),
            )


# </additional-code>



from __future__ import annotations

import dataclasses
import pathlib
import typing as t

from uniserde import JsonDoc

import rio

from . import utils
from .color import Color
from .fills import ImageFill, LinearGradientFill, SolidFill
from .self_serializing import SelfSerializing

__all__ = [
    "Font",
    "TextStyle",
]


_TextFill = SolidFill | LinearGradientFill | ImageFill | Color


@dataclasses.dataclass(frozen=True)
class Font(SelfSerializing):
    """
    A custom font face.

    The `Font` class lets you create custom fonts for use in your rio app. To
    instantiate a `Font`, you must pass it at least one font file. (As far as
    rio is concerned, the file format is irrelevant - all that matters is that
    a browser can display it.)


    ## Attributes

    `regular`: The regular (i.e. not bold, not italic) font file.

    `bold`: The bold font file.

    `italic`: The italic font file.

    `bold_italic`: The bold and italic font file.

    `ROBOTO`: A pre-defined font:
        [`Roboto`](https://fonts.google.com/specimen/Roboto).

    `ROBOTO_MONO`: A pre-defined font:
        [`Roboto Mono`](https://fonts.google.com/specimen/Roboto+Mono).
    """

    regular: pathlib.Path | bytes
    bold: pathlib.Path | bytes | None = None
    italic: pathlib.Path | bytes | None = None
    bold_italic: pathlib.Path | bytes | None = None

    def _serialize(self, sess: rio.Session) -> str:
        return sess._register_font(self)

    # Predefined fonts
    ROBOTO: t.ClassVar[Font]
    ROBOTO_MONO: t.ClassVar[Font]


Font.ROBOTO = Font(
    regular=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-Regular.ttf",
    bold=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-Bold.ttf",
    italic=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-Italic.ttf",
    bold_italic=utils.RIO_ASSETS_DIR / "fonts/Roboto/Roboto-BoldItalic.ttf",
)

Font.ROBOTO_MONO = Font(
    regular=utils.RIO_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Regular.ttf",
    bold=utils.RIO_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Bold.ttf",
    italic=utils.RIO_ASSETS_DIR / "fonts/Roboto Mono/RobotoMono-Italic.ttf",
    bold_italic=utils.RIO_ASSETS_DIR
    / "fonts/Roboto Mono/RobotoMono-BoldItalic.ttf",
)


@dataclasses.dataclass(frozen=True)
class TextStyle(SelfSerializing):
    """
    A collection of styling properties for text.

    This is a simple storage class that holds information about how text should
    be styled. You can use it to specify the font, fill, size, and other
    properties of text in your rio app.


    ## Attributes

    `font`: The `Font` to use for the text. When set to `None`, the default font
        for the current context (heading or regular text, etc) will be used.

    `fill`: The fill (color, gradient, etc.) for the text.

    `font_size`: The font size.

    `italic`: Whether the text is *italic* or not.

    `font_weight`: Whether the text is normal or **bold**.

    `underlined`: Whether the text is underlined or not.

    `strikethrough`: Whether the text should have a line through it.

    `all_caps`: Whether the text is transformed to ALL CAPS or not.
    """

    _: dataclasses.KW_ONLY
    font: Font | None = None
    fill: _TextFill | None = None
    font_size: float = 1.0
    italic: bool = False
    font_weight: t.Literal["normal", "bold"] = "normal"
    underlined: bool = False
    strikethrough: bool = False
    all_caps: bool = False

    def replace(
        self,
        *,
        font: Font | None = None,
        fill: _TextFill | None | utils.NotGiven = utils.NOT_GIVEN,
        font_size: float | None = None,
        italic: bool | None = None,
        font_weight: t.Literal["normal", "bold"] | None = None,
        underlined: bool | None = None,
        strikethrough: bool | None = None,
        all_caps: bool | None = None,
    ) -> TextStyle:
        """
        Returns an updated copy of the style.

        This function allows you to create a new `TextStyle` object with the
        same properties as this one, except for the ones you specify.

        ## Attributes

        `font`: The `Font` to use for the text. When set to `None`, the default font
            for the current context (heading or regular text, etc) will be used.

        `fill`: The fill (color, gradient, etc.) for the text.

        `font_size`: The font size.

        `italic`: Whether the text is *italic* or not.

        `font_weight`: Whether the text is normal or **bold**.

        `underlined`: Whether the text is underlined or not.

        `strikethrough`: Whether the text should have a line through it.

        `all_caps`: Whether the text is transformed to ALL CAPS or not.
        """
        return type(self)(
            font=self.font if font is None else font,
            fill=self.fill if isinstance(fill, utils.NotGiven) else fill,
            font_size=self.font_size if font_size is None else font_size,
            italic=self.italic if italic is None else italic,
            font_weight=(
                self.font_weight if font_weight is None else font_weight
            ),
            underlined=self.underlined if underlined is None else underlined,
            strikethrough=(
                self.strikethrough if strikethrough is None else strikethrough
            ),
            all_caps=self.all_caps if all_caps is None else all_caps,
        )

    def _serialize(self, sess: rio.Session) -> JsonDoc:
        return {
            "fontName": None
            if self.font is None
            else self.font._serialize(sess),
            "fill": sess._serialize_fill(self.fill),
            "fontSize": self.font_size,
            "italic": self.italic,
            "fontWeight": self.font_weight,
            "underlined": self.underlined,
            "strikethrough": self.strikethrough,
            "allCaps": self.all_caps,
        }



import asyncio

import rio
from tests.utils.layouting import BrowserClient


async def test_dropdowns_work_in_dev_tools() -> None:
    # Dropdowns (and other popups) have often been broken in the dev tools, due
    # to z-index issues and other reasons. This test makes sure that they work.

    async with BrowserClient(rio.Spacer, debug_mode=True) as client:
        # Click the 2nd entry in the sidebar, which is the "Icons" tab
        await client.execute_js(
            "document.querySelector('.rio-switcher-bar-option:nth-child(2) .rio-switcher-bar-icon').click()",
        )
        await asyncio.sleep(0.5)

        # Click an arbitrary icon in the list
        await client.execute_js(
            "document.querySelector('.rio-switcher .rio-icon').click()"
        )
        await asyncio.sleep(0.5)

        # Open the dropdown
        await client.execute_js(
            "document.querySelector('.rio-dropdown input').focus()"
        )
        await asyncio.sleep(0.5)

        # Make sure the dropdown list is open and visible
        is_entirely_visible = await client.execute_js("""
            new Promise((resolve) => {
                let elem = document.querySelector(".rio-dropdown-options");
                
                let observer = new IntersectionObserver((entries) => {
                    resolve(entries[0].intersectionRatio === 1);
                });
                observer.observe(elem);
            });
        """)
        assert is_entirely_visible
